if type(SaveGameVersion) ~= "number" or SaveGameVersion < 17 then
	error("burden of knowledge requires grimrock version 2.2.4 or later\nplease install the latest version of grimrock 2")
end

if config.editor then dofile("mod_assets/editorMod_noexport.lua") end

local ladderdotlua = [[
for i,arg in ipairs(sys.args()) do
	if arg == "burdenOfKnowledge" then
		local orig_config_load = Config.load
		Config.load = function(self)
			orig_config_load(self)
			self.renderEngine = "opengl"
			self.renderingQuality = 2
		end
		break
	end
end

dofile("Ladder.lua")]]

function burdenOfKnowledgeRestart()
	-- make the magic file that sets render engine and quality
	-- Ladder.lua is far from the only file that works for this; anything
	-- grimrock require()s after Config but before creating engineSys will
	-- work.
	local ok, err = pcall(function()
		local wd = FileSystem.getCurrentDirectory()
		local libDirectory = wd.."/lib"
		local filename = libDirectory.."/Ladder.lua"
		if FileSystem.fileExists(filename) and FileSystem.readFile(filename) == ladderdotlua then
			-- no need to write it again!
			return
		end
		if not FileSystem.fileExists(libDirectory) then
			FileSystem.createPath(libDirectory)
		end
		local file = io.open(filename,"wb")
		file:write(ladderdotlua)
		file:close()
	end)
		
	if not ok then
		if config.renderEngine == "opengl" then
			error("burden of knowledge failed to automatically change rendering quality\nselect High rendering quality and start the mod again\nsorrr\nerror details:"..err)
		else
			error("burden of knowledge failed to create necessary files in your grimrock 2 installation folder\ntry running as administrator\nsorry\nerror details:"..err)
		end
	end

	-- make a temp.sav that starts burden of knowledge
	-- while we could just do this in the lua file above and skip loading
	-- screens, a temporary save is vanilla's way of restarting after changing
	-- settings, and I'm not going to rock that particular boat right now.
	local file = NativeSaveGameOutputStream.create(SaveGameVersion)
	file:openFile(config.documentsFolder.."/temp.sav")
	
	file:openChunk("GUID")
	local modInfo = modSystem:getCurrentMod()
	file:writeValue(modInfo.guid)
	file:writeValue(modInfo.saveGameVersion)
	file:closeChunk()
	
	file:openChunk("DESC")
	file:writeValue(modInfo.name)
	file:writeValue(1) -- level
	file:writeValue(os.time())
	file:writeValue("no:time:left") -- playtime
	file:writeValue("temp") -- description
	for i=1,4 do file:writeValue(0) end -- champion levels
	file:writeValue("nowhere") -- map name
	file:closeChunk()
	
	-- we need a map with a party with the id "party"
	local saveMap = Map.create(dungeon,2,1,1)
	-- yes, this is a duplicate identity id. no, that doesn't error. no, you
	-- shouldn't try this at home.
	local saveParty = spawn(saveMap,"party",0,0,0,0,"party")
	file:openChunk("LEVL")
	file:writeValue(1)
	
	-- map:saveState() would normally be called here but we really don't
	-- need to bother with all the tiles and crap
	-- we only need the level number, visited (it expects a value there),
	-- cell bits, tiles, and the party object
	file:writeValue(false) -- visited
	
	-- ...but writing the width/height instead of the default 32x32 allows
	-- writing far fewer cell bits for a far smaller file, so i do that
	file:openChunk("SIZE")
	file:writeValue(1)
	file:writeValue(1)
	file:closeChunk()
	
	file:openChunk("CELL")
	file:writeValue(0)
	file:closeChunk()
	
	file:openChunk("TMAP")
	file:writeValue("")
	file:closeChunk()
	
	file:openChunk("ENTS")
	file:writeValue(1)
		file:openChunk("GOBJ")
		file:writeValue(1)
		saveParty:saveState(file)
		file:closeChunk() -- GOBJ
	file:closeChunk() -- ENTS
	
	file:closeChunk() -- LEVL
	
	messageSystem:delayedFunctionCall("dofile",0,"mod_assets/startFromSave.lua")
	file:openChunk("MSGS")
	messageSystem:saveState(file)
	file:closeChunk()

	-- usually saves have a lot of other chunks, but none of those are needed here
	
	file:close()
	
	sys.restart({"skipSplash","restoreGame","burdenOfKnowledge"})
end

local orig_renderLoadingScreen = renderLoadingScreen
renderLoadingScreen = function(fade)
	return orig_renderLoadingScreen(0)
end

-- force high rendering quality and opengl, because i am lazy. also we REALLY
-- need lights.
--
-- yes i'm evil and don't prompt the player about what's going to happen :p
if (config.renderingQuality == 1 or config.renderEngine ~= "opengl") and not config.editor then
	burdenOfKnowledgeRestart()
else
	dofile("mod_assets/i.lua")
end
