-- #NOEXPORT
-- We don't care to do actual editing with the editor for this dungeon, but the
-- fast loading and previewing is handy, and we might as well export it from the
-- editor instead of from some external program.
if editorModApplied then return end
editorModApplied = true

-- First off, let's not accidentally overwrite our dungeon.lua or .dungeon_editor.
function DungeonEditor:saveDungeon() end

-- This is called with no arguments to get the generated dungeon file as a
-- string. Give it the uh, not generated one.
function DungeonEditor:saveDungeonFile(filename)
	if not filename then
		return FileSystem.readFile("mod_assets/editorDungeon.lua")
	end
end



local isWindows = sys.platform() == "win32"

dungeonEditor.zhandulMenu = FrameMenu:create()
dungeonEditor.zhandulMenu:addItem(22,"Restart Dungeon Editor")
dungeonEditor.zhandulMenu:addItem(23,"Restart Dungeon Editor (OpenGL)")
if config.renderEngine == "opengl" and config.renderingQuality == 2 then
	dungeonEditor.zhandulMenu:enableItem(23,false)
end
dungeonEditor.zhandulMenu:addItem(24,"Restart Dungeon Editor (Direct3D9)")
if isWindows and config.renderEngine == "d3d9" and config.renderingQuality == 2 then
	dungeonEditor.zhandulMenu:enableItem(24,false)
end
dungeonEditor.zhandulMenu:addItem(25,"Restart Dungeon Editor (OpenGL, notebook)")
if config.renderEngine == "opengl" and config.renderingQuality == 1 then
	dungeonEditor.zhandulMenu:enableItem(25,false)
end
dungeonEditor.zhandulMenu:addItem(26,"Restart Dungeon Editor (Direct3D9, notebook)")
if isWindows and config.renderEngine == "d3d9" and config.renderingQuality == 1 then
	dungeonEditor.zhandulMenu:enableItem(26,false)
end
dungeonEditor.zhandulMenu:addSeparator()
dungeonEditor.zhandulMenu:addItem(27,"Recompile Shaders")
dungeonEditor.menuBar:addSubMenu(dungeonEditor.zhandulMenu,"Zhandul")
mainFrame:setMenuBar(dungeonEditor.menuBar)

local old_onMenuEvent = DungeonEditor.onMenuEvent
DungeonEditor.onMenuEvent = function(self, event)
	old_onMenuEvent(self,event)
	if self.dialog then return end
	if event >= 22 and event <= 26 then -- restart editor
		if event == 23 or event == 24 then
			if config.renderingQuality ~= 0 then
				config.renderingQuality = 2
				config:save()
			end
		elseif event == 25 or event == 26 then
			if config.renderingQuality ~= 1 then
				config.renderingQuality = 1
				config:save()
			end
		end
		if event == 23 or event == 25 then
			if config.renderEngine ~= "opengl" then
				config.renderEngine = "opengl"
				config:save()
			end
		elseif event == 24 or event == 26 then
			if config.renderEngine ~= "d3d9" then
				config.renderEngine = "d3d9"
				config:save()
			end		
		end
		self:onRestartDungeonEditor()
	elseif event == 27 then -- recompile shaders
		-- TODO
	end
end

-- XXX: doesn't ask to confirm closing the project. I never want it to, of course,
-- but it's inconsistent with the rest of the options that close the project.
DungeonEditor.onRestartDungeonEditor = function(self)
	self:closeProject()
	mainFrame:setMenuBar(nil) -- C function expects an explicit nil...what the hell
	sys.restart({"launchEditor"})
end

--------------------------------------------
-- more useful ProjectileCollider debug draw
--------------------------------------------
function ProjectileColliderComponent:update()
--[[
	local re = self.node:getRenderEntity()
	local hidden = re:getHidden()
	if self.enabled then
		if hidden then
			re:setMaterial(dungeon.materials.color_add.material)
			re:setEmissiveColor(vec(0.5,0,0.5))
			re:setHidden(false)
		end
	elseif not hidden then
		re:setHidden(true)
	end
	]]
	
	if self.enabled and (self.debugDraw or config.drawColliders) then
		local pos = self.node:getWorldPosition()
		pos[2] = party.go:getWorldPosition()[2]+party.eyeLevel
		DebugDraw.drawSphere(pos, 0.3, Color.White)
	end
end

-- lol
dofile("zhandulUmod/preprocessor.lua")
dofile = function(filename)
	local f,err = loadfile_withPreprocessor(filename)
	if not f then error(err) else return f() end
end

-- Obviously we need very different export behavior from normal...
function DungeonEditor:exportDungeon(filename, progressCallback)
	-- no entity/script validation step

	------------------------
	-- vanilla setup code
	------------------------
	if not filename then
		-- prompt for filename
		local params = {
			parent = mainFrame,
			type = "save",			
			description = "Dat files",
			extension = "dat",
			title = "Export Dungeon As",
		}

		if sys.platform() == "win32" then
			-- on Win7 initialDir does not always work:
			-- "If lpstrInitialDir has the same value as was passed the first time the application
			-- used an Open or Save As dialog box, the path most recently selected by the user is
			-- used as the initial directory."
			-- to workaround this we'll set initialName to a full path
			local path,filename,ext = FileSystem.splitPath(self.filename)
			params.initialName = config.documentsFolder.."/Dungeons/"..filename..".dat"
		else
			local path,filename,ext = FileSystem.splitPath(self.filename)
			params.initialDir = config.documentsFolder.."/Dungeons"
			params.initialName = filename..".dat"
		end

		filename = sys.fileDialog(params)
		if not filename then return end
	end
	
	local status,rv = pcall(FileOutputStream.create, filename)
	if not status then
		local err = rv
		self:messageBox("Export Failed", rv)
		return
	end
	
	local archive = rv
	------------------------
	-- end vanilla setup code
	------------------------
	
	
	-- collect files to export
	-- no extension restrictions here, anything in mod_assets without "noexport" gets exported
	local path = FileSystem.getPath(self.filename).."/mod_assets"
	local files = FileSystem.findFiles(path, "*", true)
	
	local i = 1
	while i < #files do
		if files[i]:find("noexport") then
			table.remove(files,i)
		else
			i = i+1
		end
	end
		
	local totalFiles = #files + 1	-- +1 for manifest
	
	-------------------------------
	-- vanilla header+manifest code
	-------------------------------
	-- header
	archive:writeBytes("GRA2")
	archive:writeInt(totalFiles)
	
	-- make space for file list
	local fileListPos = archive:getPosition()
	for i=1,totalFiles do
		archive:writeInt(0) -- filename hash
		archive:writeInt(0) -- file offset
		archive:writeInt(0) -- compressed size
		archive:writeInt(0) -- uncompressed size
		archive:writeInt(0)	-- options
	end

	local fileInfos = {}
	
	-- export dungeon manifest (save game version, name, author, description)
	do
		local fileInfo = {}
		fileInfo.hash = sys.fnv32("mod_assets/dungeon_manifest.bin")
		fileInfo.fileOffset = archive:getPosition()
		fileInfo.compressedSize = 0
		fileInfo.options = 0
		
		local dungeonInfo = dungeon.dungeonInfo		
		
		-- generate save version identifier
		local saveVersion = sys.createGuid()
		archive:writeInt(#saveVersion)
		archive:writeBytes(saveVersion)
		
		archive:writeInt(#dungeonInfo.name)
		archive:writeBytes(dungeonInfo.name)
		archive:writeInt(#dungeonInfo.author)
		archive:writeBytes(dungeonInfo.author)
		archive:writeInt(#dungeonInfo.description)
		archive:writeBytes(dungeonInfo.description)
		archive:writeInt(#dungeonInfo.dungeonFolder)
		archive:writeBytes(dungeonInfo.dungeonFolder)
		
		fileInfo.uncompressedSize = archive:getPosition() - fileInfo.fileOffset

		-- store file info
		fileInfos[#fileInfos+1] = fileInfo
	end
	-----------------------------------
	-- end vanilla header+manifest code
	-----------------------------------
	
	for i=1,#files do
		local source = FileSystem.readFile(files[i])
		local destFile = files[i]:sub(#FileSystem.getPath(self.filename) + 2)
		
		--print(files[i].." -> "..destFile)

		local fileInfo = {}
		fileInfo.hash = sys.fnv32(destFile)
		fileInfo.fileOffset = archive:getPosition()
		
		local bytes = source
		
		local ext = FileSystem.getFileExtension(files[i])

		if ext == "lua" then
			-- big changes to exporting lua files; obviously it needs to run
			-- my preprocessor before export, but I've also made it export
			-- bytecode instead of source.
			local chunk,err = loadstring_withPreprocessor(source, "@"..destFile)
			if not chunk then
				self:messageBox("Export Failed", "Error in "..destFile..":\n"..err)
				archive:dispose()
				return
			end
			
			-- export non-stripped bytecode
			-- grimrock's luajit version is too old to support
			-- deterministic bytecode generation via string_dump...
			bytes = string_dump(chunk, false)
		elseif ext == "h" then
			-- HACK: obviously, ".." doesn't work in an archive file system, but i like
			-- being able to #include from mod_assets, so...
			fileInfo.hash = sys.fnv32("shaders/gl/../../"..destFile)
		end
		
		-----------------------------------------
		-- the rest of this loop is vanilla code
		----------------------------------------
		fileInfo.uncompressedSize = #bytes

		-- compress file
		bytes = sys.compress(bytes)
		fileInfo.compressedSize = #bytes
		fileInfo.options = 1	-- compressed file

		-- write file to archive
		archive:writeBytes(bytes)

		-- store file info
		fileInfos[#fileInfos+1] = fileInfo
		
		--print(string.format("hash=%08x  offset=%d  compressed=%d  uncompressed=%d  ratio=%.1f", 
		--	fileInfo.hash, fileInfo.fileOffset, fileInfo.compressedSize, fileInfo.uncompressedSize, 
		--	fileInfo.compressedSize / fileInfo.uncompressedSize * 100))
		
		if progressCallback then progressCallback() end
	end
	
	------------------------------------
	-- vanilla code only past this point
	------------------------------------
	assert(#fileInfos == totalFiles)
	
	-- write file list to beginning
	archive:seek(fileListPos)
	for i=1,#fileInfos do
		local fileInfo = fileInfos[i]
		archive:writeInt(fileInfo.hash)
		archive:writeInt(fileInfo.fileOffset)
		archive:writeInt(fileInfo.compressedSize)
		archive:writeInt(fileInfo.uncompressedSize)
		archive:writeInt(fileInfo.options)
	end
	
	archive:dispose()
	
	return true
end
