#version 150
#include "base.h"

in vec4 position;
in vec3 normal;
#ifdef NORMAL_MAP
in vec4 tangent;
#endif
in vec2 texcoord;
#ifdef SKINNING
in vec4 boneIndices;
in vec4 boneWeights;
#endif

out vec3 positionVarying;
out vec3 normalVarying;
#ifdef NORMAL_MAP
out vec3 tangentVarying;
out vec3 bitangentVarying;
#endif
out vec2 texcoordVarying;

uniform mat4 g_modelView;
uniform mat4 g_proj;

uniform float g_time;

void main()
{
	vec4 vposition = position;
	vec3 vnormal = unpackVertexNormal(normal);
#ifdef NORMAL_MAP
	vec4 vtangent = unpackTangent(tangent);
#endif
	vec2 vtexcoord = unpackTexcoord(texcoord);

	// skinning
#ifdef SKINNING
	mat4 skinMat = skinningMatrix(boneIndices, boneWeights);
	vposition = vposition * skinMat;
	vnormal = vec3(vec4(vnormal, 0.0) * skinMat);
#ifdef NORMAL_MAP
	vtangent.xyz = vec3(vec4(vtangent.xyz, 0.0) * skinMat);
#endif
#endif

	float time = g_time+vtexcoord.x*1000.0;

	// TODO: better movement model, make it look like a breeze
	vposition.x += sin(time)*vtexcoord.y*0.01;
	vposition.y += cos(time)*vtexcoord.y*0.01;

	vec4 pos = g_modelView * vposition;

	gl_Position = g_proj * pos;
	positionVarying = pos.xyz;
	normalVarying = vec3(g_modelView * vec4(vnormal, 0.0));
#ifdef NORMAL_MAP
	tangentVarying = vec3(g_modelView * vec4(vtangent.xyz, 0.0));
	vec3 bitangent = cross(vnormal.xyz, vtangent.xyz) * vtangent.w;	// tangent.w = handedness
	bitangentVarying = vec3(g_modelView * vec4(bitangent, 0.0));
#endif
	texcoordVarying = vtexcoord;
}
