-- by may
-- 1.0.1

-- changelog:
-- 1.0.1:
--   - added an error message for trying to run the mod with grimrock versions that are too old
-- 1.0.0:
--   - initial release

--[[
a small, game-jam-esque puzzle mod i made intermittently during january 2025 and the last few days of december 2024. it is inspired by my inability to deal with tragedy, and also by a weird dream my girlfriend had involving lights.

features:
- no new classes, races, skills, traits, items, or monsters
- source code is not available yet
- you can't customize your controls
- probably incompatible with 7soul's umods
- can be finished in three seconds
- despite the short development time, easily the most meaningful thing i've made so far

i gave myself a restriction making this mod: only 128 pixels of texture data are used for non-font purposes. this resulted in complicated models and a weird aesthetic.

download: <http://cowmuffins.net/grimrock/2/burden_of_knowledge_1_0_1.zip>

this should work with all official grimrock 2.2.4 versions at time of writing (windows, mac, nutcracker beta versions), but **it will crash on startup if your grimrock 2 installation folder is write-protected (unless you're on the mac version *and* have high rendering quality selected)**. no, i probably won't ever find a workaround for that.
this mod does not modify your existing game files, it just creates a new lua file and a new savegame file.
]]

#define MATH_PI 3.141592653589793115997963468544185161590576171875

dofile("mod_assets/zhLib/mat.lua")
dofile("mod_assets/zhLib/quat.lua")
dofile("mod_assets/zhLib/sound.lua")
dofile("mod_assets/zhLib/tween.lua")
dofile("mod_assets/zhLib/vec.lua")
dofile("mod_assets/zhLib/particles.lua")
dofile("mod_assets/sounds.lua")

local currentMouseMode = "normal"

local nop = function() end
local function findObject(id)
	return dungeon.maps[1]:findEntity(id)
end

local GameMode = GameMode

-- optimization upvalues
-- not all of these are necessarily productive since not all of them are used
-- often but it's easy ;p
local mat_create = mat.create
local mat_mul = mat.mul
local mat_rotationAbout = mat.rotationAbout
local mat_rotationX = mat.rotationX
local mat_rotationY = mat.rotationY
local mat_rotationZ = mat.rotationZ
local mat_scaling = mat.scaling
local mat_toDirectionVector = mat.toDirectionVector
local mat_transformVector = mat.transformVector
local mat_transformVectorNum = mat.transformVectorNum

local math_ceil = math.ceil
local math_clamp = math.clamp
local math_cos = math.cos
local math_floor = math.floor
local math_max = math.max
local math_min = math.min
local math_rad = math.rad
local math_random = math.random
local math_randomseed = math.randomseed
local math_sin = math.sin

local sys_keyDown = sys.keyDown
local sys_mousePos = sys.mousePos

local vec_addInPlaceV = vec.addInPlaceV
local vec_cross = vec.cross
local vec_length3 = vec.length3
local vec_lengthSquared = vec.lengthSquared
local vec_mulInPlace = vec.mulInPlace
local vec_normalize = vec.normalize
local vec_normalizeInPlace3 = vec.normalizeInPlace3

local tempRay = {}
local tempBox = {}

#define POINT_MODEL_FILE "mod_assets/point.fbx"

-- The placement of Linwfr's name is optimized for these, so if you change them,
-- change that too.
#define ROOM_SIZE 9
#define ROOM_WALL_MARGIN 1

local function infoMessage(msg)
	for i,line in ipairs(msg:split("\n")) do gui:hudPrint(line) end
end

---------------------------------------------
-- Begin goddess data
---------------------------------------------
--
-- Do NOT change a single thing here, it's puzzle-critical info. Also since the
-- preprocessor is primitive and I'm lazy sometimes I typed their names directly
-- into strings anyway.
--
---------------------------------------------

-- first: given outright
-- physicality (needed for narrative)
-- number: 7 e
#define PHYS_NAME Frilwn
#define PHYS_NAME_LOWER frilwn
#define PHYS_NUMBER 7

-- second: given outright
-- water
-- number: 4 y
#define WATER_NAME Irsten
#define WATER_NAME_LOWER irsten
#define WATER_NUMBER 4

-- third: after unsealing the ceiling, look at the stars from the right
--   perspective and they spell out the name
-- earth (needed for easy placement of filter for Fej*sikomb text)
-- number: 2 a
#define EARTH_NAME Linwfr
#define EARTH_NAME_LOWER linwfr
#define EARTH_NUMBER 2

-- fourth: after learning to zoom in, look at tiny text through a filter
--   held by Linwfr or Frilwn (Frilwn is cooler but Linwfr is easier)
-- fire
-- number: 3 s
#define FIRE_NAME Fej*sikomb
#define FIRE_NAME_LOWER fej*sikomb
#define FIRE_NUMBER 3

-- fifth: draw a pentagram via incantation, starting at A
-- air
-- number: 5 *
#define AIR_NAME Azchxa
#define AIR_NAME_LOWER azchxa
#define AIR_NUMBER 5

-- sixth:
-- life and death
-- number: 1 M and 6 m
#define LIFE_NAME Pyiqe
#define LIFE_NAME_LOWER pyiqe
#define DEATH_NAME Kvotomu
#define DEATH_NAME_LOWER kvotomu
#define LIFE_NUMBER 1
#define DEATH_NUMBER 6

-- seventh: The shapes drawn on the floor constellations when you incant the
-- previous goddess's names are the letters of this one.
--
-- "Tell the stars seven names to see the eighth"
-- number: 0
#define FINAL_NAME Masy*me
#define FINAL_NAME_LOWER masy*me
#define FINAL_NUMBER 9

#define APPROX_GODDESS_HEIGHT 0.8

local PHYS_COLOR = vec(1,1,1) -- Frilwn
local WATER_COLOR = vec(0,0,1) -- Irsten
local EARTH_COLOR = vec(0,1,0) -- Linwfr
local AIR_COLOR = vec(0,0,-1) -- Azchxa
local FIRE_COLOR = vec(0,-1,0) -- Fej*sikomb
local LIFE_AND_DEATH_COLOR = vec(1,0,0) -- Pyiqe and Kvotomu
local FINAL_COLOR = vec(-1,-1,-1)

local function goddessPosition(vertexIndex)
	local PEDESTAL_CENTER = vec(4.5, PEDESTAL_HEIGHT, 7.5)
	local vertexCount = 12
	local radius = 2.3
	local angle = vertexIndex/vertexCount*MATH_PI*2
	return PEDESTAL_CENTER+vec(math_sin(angle), 0, math_cos(angle))*radius
end

local goddessData = {
	[LIFE_NUMBER] = {
		color = LIFE_AND_DEATH_COLOR,
		mesh = "mod_assets/gld.fbx",
		position = goddessPosition(7.5),
		id = "goddess_life_death",
	},
	[DEATH_NUMBER] = {
		color = LIFE_AND_DEATH_COLOR,
		mesh = "mod_assets/gld.fbx",
		position = goddessPosition(7.5),
		id = "goddess_life_death",
	},
	[EARTH_NUMBER] = {
		color = EARTH_COLOR,
		mesh = "mod_assets/ge.fbx",
		position = goddessPosition(3.5),
		id = "goddess_earth",
	},
	[FIRE_NUMBER] = {
		color = FIRE_COLOR,
		mesh = "mod_assets/gf.fbx",
		position = goddessPosition(4.5),
		id = "goddess_fire",
	},
	[WATER_NUMBER] = {
		color = WATER_COLOR,
		mesh = "mod_assets/gw.fbx",
		position = goddessPosition(6.5),
		id = "goddess_water",
	},
	[AIR_NUMBER] = {
		color = AIR_COLOR,
		mesh = "mod_assets/ga.fbx",
		position = goddessPosition(8.5),
		id = "goddess_air",
	},
	[PHYS_NUMBER] = {
		color = PHYS_COLOR,
		mesh = "mod_assets/gp.fbx",
		position = goddessPosition(5.5),
		id = "goddess_physicality",
	},
	[FINAL_NUMBER] = {
		color = FINAL_COLOR,
		mesh = "mod_assets/gfinal.fbx",
		position = vec(4.5,1.8,3.5),
		id = "goddess_final",
	},
}

---------------------------------------------
-- End goddess data
---------------------------------------------

#define PEDESTAL_HEIGHT 0.8

#define CEILING_DESTRUCTION_INCANTATION unsealing
#define FINAL_INCANTATION indifferencekillsus

-- these are manually spaced so if you change their length you also have to change
-- gui:updateHudPrints() down there...
#define MUTE_INCANTATION silence
#define UNMUTE_INCANTATION listen

local bookCoverText = [[
proceed only if you would
bear the burden of knowledge

this is the only warning
the last thing you will ignore
]]

--------------------------------------------------------------------------------
-- first
--------------------------------------------------------------------------------

local bookTextUnlit = [[
The first actor of new creation     
Her light leading to all others     
Evoke her image by speaking her name

Frilwn                              ]]

--------------------------------------------------------------------------------
-- second
--------------------------------------------------------------------------------

local bookTextWhite = [[
Incantations are keys to understanding.
Numbers are keys to understanding.     
A number exists for every name.        
Lifeless images may still move.        

Incant her number and see.             ]]

local bookTextWhiteSmall = [[
from a to z and four more lines
her name paints her star
}
]]

--------------------------------------------------------------------------------
-- third
--------------------------------------------------------------------------------

-- pattern on the ceiling should hint that you should light it up
local bookTextBlue = [[
No pattern repeats forever.        
Ceilings on progress are false.    
A way up exists; stop looking down.
Navigate those stars to the future.
Turn and find your freedom.        
]]

local bookTextBlueSmall = [[
                         *        
       *                          
    forever, what a joke          
 *  we won't last 'til next week  
    no maybes, it's too late     *
                   *              
            *            *        
]]

--------------------------------------------------------------------------------
-- fourth
--------------------------------------------------------------------------------

local bookTextGreen = [[
A word cannot act.                       
Their power comes from you.              
Interpretation means attention to detail.
Only close observation will find truth.  

         +                   $%$         
]]

local bookTextGreenSmall = [[
EARTH_NAME_LOWER's ~ sees the night
by WATER_NAME_LOWER's light
]]

local bookTextCyan = [[
Several texts exist.                

Use your power of close observation.
Some contain further meaning.       

      +                   $%$      
]]

local bookTextCyanSmall = [[
read by the light of EARTH_NAME_LOWER 
to find she whose name contains a star
read by the light of PHYS_NAME_LOWER  
to find she whose name creates a star 
]]

--------------------------------------------------------------------------------
-- requires -green but not -blue
--
-- at this point, things open up a lot, as you can invoke Azchxa and Fej*sikomb
-- in either order.
--------------------------------------------------------------------------------

-- extra hint for Azchxa, who you may not have yet
local bookTextMagenta = [[
Negation of light is not darkness.   
Darkness teaches nothing.            
Invoke she whose name draws her star.
Form her star with five lines.       
From a to z is the first.            
}
]]

local bookTextNGreen = [[
there is nothing outside         
for the walls to protect you from
]]

local bookTextBlueNGreen = [[
   #      #   #  #          
first of the inseparable two
help create starry truth    
 she is here imbricated     
 search in lines quilted    
look here, for she is here  
]]

--------------------------------------------------------------------------------
-- requires -blue but not -green
--------------------------------------------------------------------------------
local bookTextYellow = [[
Endings need not be beginnings.
Reader, you near that ending.  
Escape is a fantasy.           
Nothing awaits after the end.  
]]

-- extra hint for Fej*sikomb, who you may not have yet
local bookTextNBlue = [[
negation of light is not darkness       
darkness teaches nothing                
invoke the name hiding in the blue night
see that night through EARTH_NAME_LOWER's ~       
]]

local bookTextGreenNBlue = [[
one of the inseparable two
123435u
]]

--------------------------------------------------------------------------------
-- requires -blue and -green
--------------------------------------------------------------------------------
local bookTextRed = [[
No one reads alone.                
              
Isolation is impossible.           
Seek the inseparable two.          
   
Invoke both, lest you find neither.
]]

local bookTextNCyan = [[
everything cannot be saved
]]

--------------------------------------------------------------------------------
-- requires red
--------------------------------------------------------------------------------
local bookTextRedNGreen = [[
speak seven names  
to see the last    
their seven numbers
show the order     
]]
local bookTextRedNBlue = bookTextRedNGreen
local bookTextRedNCyan = bookTextRedNGreen
local bookTextYellowNBlue = bookTextRedNGreen
local bookTextMagentaNGreen = bookTextRedNGreen
-------------------
-- requires masy*me
-------------------

local bookTextFinal = [[
|                                     
Columns hold the final incantation.   
Evoke its power and rise.             

Keep no knowledge from the last world.
It would only mislead.                
Lifeless stars will remain.           
Lands with no meaning.                
{                                     
]]





local hints = {
	hint = {
		"incant 'hint' again to get a hint"
	},
	[PHYS_NUMBER] = {
		"if you figured out how to get hints, you can figure out how to evoke frilwn's image",
	},
	[WATER_NUMBER] = {
		"PHYS_NAME_LOWER's number is 7\nwhat kind of key would that be?",
		"press the 7 key on your keyboard",
		"the 7 key turns frilwn's image around, lighting up the opposite wall\none wall has text telling you the next actor's name\nincant that name, like how you incanted 'frilwn'",
	},
	ceiling = {
		"look around using the mouse\nmove using the arrow keys on your keyboard",
		"look up",
		"there is a word written on the ceiling above you\nincant that word",
	},
	[EARTH_NUMBER] = {
		"examine the stars, and look for anything strange",
		"some stars are a different color, and concentrated in certain areas",
		"try looking at the stars from a different angle",
		"get as close to the text as possible, then turn around and look up at the stars\nit might help to try moving around a bit",
		"the stars spell out 'linwfr', the next actor's name",
	},
	zoom = {
		"you have the power of close observation\nbut you haven't used it yet",
		"the icons beneath the text tell you how to use the power of close observation",
		"hold the + key on your keyboard or hold your middle mouse button",
		"some of the texts also have tiny text alongside them\nzoom in on it to read it",
	},
	fejOrAzchxa = {
		"there are four tiny texts, all in the same place\ncheck the blue, green, cyan, and white texts",
		"there are two different actors you can invoke at this point\nthe hints will start with the one whose name contains a star",
	},
	[FIRE_NUMBER] = {
		"the name of the one whose name contains a star\nis hidden within one of the tiny texts",
		"linwfr's image has something you need, but it isn't light\nthe night can only be seen in irsten's light, not linwfr's",
		"use irsten's light and turn everyone else away\nstand just behind linwfr and zoom in on the tiny blue text,\nlooking through linwfr's circle/hexagon/cube thing",
		"linwfr's circle/hexagon/cube thing contains tiny circles that circle letters in the text\nshowing the name of the one whose name contains a star: 'fej*sikomb'",
		"the star in 'fej*sikomb' is not a letter, but, well, in THIS font, it's pretty obvious\nthat there's a star on your keyboard, and you just type that as part of the incantation\nthe 8 key and the numpad one both work",
	},
	[AIR_NUMBER] = {
		"the name of the one whose name creates a star must be discerned by you\nit is not directly shown in the texts",
		"try typing some letters, and look at what happens on the floor as you do so\nif things get too cluttered, you can use backspace/delete/enter",
		"typing incantations creates lines on the floor between each pair of letters",
		"\"from a to z\" is literal\nthe name that creates a star begins with those two letters",
		"start by incanting 'az', then look near the edges for letters that will draw the star shown in the texts (a pentagram) when incanted\nit takes five lines to draw a pentagram, and you know it starts with 'az' and ends with 'a'\nso you only need three more letters",
		"the name that creates a star is 'azchxa', pronounced 'azchxa'",
	},
	lifeOrDeath = {
		"you have what you need to figure out either of the names of the inseparable two\nthe hints will start with the first of the inseparable two",
	},
	[LIFE_NUMBER] = {
		"all information about the name of the first of the inseparable two\nis in a single text, the one that appears in the light of irsten and fej*sikomb",
		"there are four small characters at the top of the text referring to the first of the inseparable two\nthey mean something important",
		"the four small characters are arrows pointing down\nread the letters they point to top-to-bottom, ignoring the rest of the text",
		"reading the four pointed-to columns of letters gives you the message \"speak her name pyiqe\"\nso do that",
	},
	[DEATH_NUMBER] = {
		"the clue for the name of the remaining member of the inseparable two\nis in the text that appears in the light of azchxa and linwfr\nbut you'll need information from elsewhere in the room, too",
		"you need to translate the stars with numbers inside them into letters\nyou don't need to translate 'u' into a letter, it already is one",
		"the stars match the stars at the feet of the various images you've already invoked\nexamine those stars",
		"a three-pointed star with the number '7' in it means the seventh letter printed on the three-pointed star in the room\nin this case, that would be the seventh letter on fej*sikomb's star, 'k'",
		"the seventh letter on azchxa's star is 'v'",
		"the name of the remaining member of the inseparable two is 'kvotomu'",
	},
	[FINAL_NUMBER] = {
		bookTextRedNGreen,
		"okay, fine, hints\nremember how azchxa's name draws a star?",
		"all seven names of the people you've invoked so far draw something",
		"the seven names draw letters (except for azchxa's, which draws a star, as we've already established)",
		"each name has a corresponding number, which you had to know to get this far in the first place\nuse those numbers to order the letters in the obvious way",
		"the last name is 'masy*me'",
	},
	final = {
		"this final puzzle is pretty easy, really\nthe colors in the text that appears in masy*me's light are the colors of other texts\nbut the puzzle can also be done while ignoring that clue",
		"masy*me can be turned around with the 9 key on your keyboard\nlook at all the texts available, keeping this last one in mind",
		"capital letters are uncommon, and always aligned in a column",
		"by ordering all the texts with capital letters according to the colors in masy*me's text\nyou will find a message",
		"the message is \"THE FINAL INCANTATION IS INDIFFERENCE KILLS US\", which should be pretty self-explanatory",
	},
}








				


----------------------------
--
-- Utility functions.
--
----------------------------

-- HACK: store last absolute mouse position. This is used so that held items can
-- continue to follow the camera during mouse look.
-- All this is unused in this particular mod
local lastMouseX = -1
local lastMouseY = -1
local function getMouseRay()
	if party.mouseLook then
		return gameMode:getWorldRay(lastMouseX, lastMouseY)
	else
		return gameMode:getWorldRay(sys_mousePos())
	end
end

local gameModeUpdate = GameMode.update
-- WARNING: this only works for keys that have names according to sys.getKeyName()
local keyStates = {}
local lastMouseModeChangeFrame = -1
function GameMode:update()
	local desiredMouseMode = (linwfrKeyEditMode or infoMode or self.menu or (config.editor and not (dungeonEditor.previewMode and dungeonEditor.fullscreen))) and "normal" or "relative"
	if currentMouseMode ~= desiredMouseMode then
		mainFrame:setMouseMotionMode(desiredMouseMode)
		currentMouseMode = desiredMouseMode
		lastMouseModeChangeFrame = Time.currentFrame
	end

	if party and not party.mouseLook then lastMouseX, lastMouseY = sys_mousePos() end -- Unused in this mod
	
	_tween_update(Time.deltaTime)
	gameModeUpdate(self)
	
	for keyName,down in pairs(keyStates) do
		if not sys_keyDown(keyName) then keyStates[keyName] = nil end
	end
	
	-- stupid place to put this...
	party.go.map:updateComponents(ZhandulParticleComponent)
end



--------------------------------------------------------------------------------
--
-- Get rid of "do you really want to quit?" prompt.
--
--------------------------------------------------------------------------------
closeWindow = function() sys.exit() end

--------------------------------------------------------------------------------
--
-- Remove all champion UI.
--
--------------------------------------------------------------------------------

-- eats mouse presses
AttackPanel.update = nop

-- draws and does a bunch of other stuff we don't want
AttackFrame.update = nop

-- efficiently gets rid of everything, cool!
CharSheet.show = nop

--------------------------------------------------------------------------------
--
-- Remove resting.
--
-- This would cause RopeToolComponent and DiggingToolComponent to produce
-- corrupted saves, but since we aren't using those components we don't care.
--
--------------------------------------------------------------------------------
PartyComponent.rest = nop

--------------------------------------------------------------------------------
--
-- Remove saving and loading; better for players not to have the option of
-- worrying about it.
--
--------------------------------------------------------------------------------
local saveAttemptMessage = "your progress is saved automatically\nthere is nothing else to save"
local loadAttemptMessage = "you cannot turn back time"
function GameMode:saveGame()
	infoMessage(saveAttemptMessage)
end
function GameMode:loadGame()
	infoMessage(loadAttemptMessage)
end
function GameMode:showMap()

end

--local debugDrawRays = {}
endMessagePage = 1
local endMessagePages = {
[[you have completed the game.

i made this over about a month, probably working about 100 hours, as a sort of unpolished game jam kind of thing.

to read the developer's commentary, press 7.

this is the only ending; you didn't miss anything.
to reset your progress anyway, delete burden_of_knowledge.sav.
it's in the same folder as your regular grimrock saved games.]],
[[Thanks for playing! The narrative came out pretty overt and hopeless, huh? Just remember, the world in real life hasn't ended yet.

I knew from the beginning that I wanted this mod to be set in the void after Grimrock's world ends, and spin that into an examination of--well, you just played it, you know the kind of things it's about. I also arrived almost instantly at the idea of a small, not-very-diverse group of women being tasked with creating a world, and the player only seeing shallow figurines of them that can't really do anything.
The writing style of the clues got kicked around a little more; in the first draft a lot more of them were implied to be direct quotes from said women.

I didn't decide what to do for the ending until near the end of development. I always knew it'd be dim - losing hope is the easy thing to do, so it's an appropriate ending here - but for most of development I didn't think I'd make it this bleak, heh.

(Press 7 for next page)]],
[[The puzzles came together quickly. The initial idea my girlfriend's dream gave me was for the figurines to be smaller objects you can pick up and move around, and for a bunch of lights in the room to change their colors and directions according to the figurines' positions. This was discarded almost immediately in favor of the figurines themselves being light sources, and moving them around was discarded soon after.

Text that's only readable under a certain color of light is an idea I've wanted to use for over 10 years. While I briefly considered trying to make an image where this arises naturally with normal lighting behavior, I went with the completely-invisible-except-in-matching-light behavior instead because it's usable by colorblind players and gives me an opportunity to play with negative lighting. Definitely not just because it's easier, nope.

The big challenge with the puzzle design, of course, was the layout of the letters on the floor. I wanted all the names to look almost pronouncable, and I needed all of those names except Masy*me's and Azchxa's to trace the shapes of letters on the floor, and Azchxa's letters have to be far apart from the rest so the star is cleanly identifiable...it took some work to get a set of points and names where everything sorta functions.

Surprising nobody, Masy*me's name is definitely my favorite puzzle. Funny how the actual final puzzle is so lame by comparison.

(Press 7 for next page)]],
[[At the beginning of development, I was going to try to fit this mod into 64K. That got tossed aside in favor of a different restriction: avoiding textures and texture-like effects where possible, except for fonts. The aesthetic of the mod flowed naturally from that, but don't expect most of it to return - I'm surprised with how much I like the look of the super-primitive dithering lighting, and I like pixel particles in anything, but the rest of the low-color outliney visual style I wouldn't be fond of even if it was done properly.

Even though I had a nice premade rigged model from the Internet to start with, making the models for the characters was by far the biggest workload of the project. Except for Kvotomu and Masy*me, the poses and accompanying hair all had to do something very specific. I learned Blender's "new" (added in like 3.4) hair curves system for this purpose, which was a big help but I still had a lot of difficulty (it turns out that since real people don't typically have hair that's several meters long and prehensile, a lot of 3D modeling tools aren't designed for making that). You can probably tell I modeled Frilwn first.
If this had been a bigger, more serious project, I would've had to optimize the hair to not use hundreds of thousands of triangles...geez that'd be miserable.
The part that really embarrasses me is that everyone's arms and legs are identical - sure the figurines are meant to be unsettlingly similar but not THAT similar!

As usual, the code doesn't do anything new or impressive. I spent a lot of time chasing a dead-end trying to do memory hacks to use custom shaders with Grimrock's native particle system code, or at least modify meshes faster, with nothing to show for it (would've been a compatibility nightmare anyway).

The from-scratch custom particle system implementation was written specifically for this mod, despite being named after Zhandul's Dream. I also found out the game really doesn't like it when I assign tangents to a mesh that don't match its geometry.

Because the whole thing's hacked together as fast as possible (remember, unpolished game jam kind of thing), and because a lot of traces of the 64K-targeting code remain, the code quality here is absolutely horrendous. Everyone says their code is a mess, but this really is the most horrible mess you could imagine. Most of it is in a single file!

(Press 7 for next page)]],
[[The sound was done in one day near the end of development. I forced myself NOT to put effort into it, since nothing else was going to be polished either. I just used off-the-shelf background ambience, and the sound effects are equally slapdash.

The sound effect of the ceiling being destroyed was my first time using horrible dynamic range compression on purpose. I still feel dirty about that, even though approximately every other game/movie/etc. does it. Or maybe that's why I feel dirty about it.

Don't worry (or do), I'm sure my next mod will go back to using tracker music. Will probably outright use libopenmpt or BASS or something.

Anyway I love my girlfriend I love being gay

(Press 7 for next page)]],
1,
}
local bonusTex = RenderableTexture.load("mod_assets/bonus.tga")
gui:reset()
local orig_updateHudPrints = Gui.updateHudPrints
-- also this is like just a convenient place to hook into sorry
function Gui:updateHudPrints()
	if infoMode then
		local instructions = string.format([[
esc or click           resume mouse control
enter                  clear
alt+f4                 quit %s
]],saveFailed and "(you will lose your progress)" or "(you won't lose progress)")
		if config.muteMusic then
			instructions = "incant 'UNMUTE_INCANTATION'        hear us again\n"..instructions
		else
			instructions = "incant 'MUTE_INCANTATION'       be deafened\n"..instructions
		end
		local font = FontType.Default
		local _,height = ImmediateMode.drawParagraph(instructions, -10000, -10000, 1000, font, Color.White)
		ImmediateMode.drawParagraph(instructions, 4, config.height-4-height, 1000, font, Color.White)
	end
	
	if gameDone then
		local message = endMessagePages[endMessagePage]
		
		if message == 1 then
			local w = bonusTex:getWidth()
			local h = bonusTex:getHeight()
			ImmediateMode.drawImage(bonusTex, config.width/2-w/2, config.height/2-h/2, w, h, w, h, Color.White)
		else
			local font = FontType.Default
			local maxwidth = math_min(config.width-128,1024)
			local width,height = ImmediateMode.drawParagraph(message, -10000, -10000, maxwidth, font, Color.White)
			ImmediateMode.drawParagraph(message, config.width/2-width/2, config.height/2-height/2, maxwidth, font, Color.White)
		end
	else
		-- draw hudprints but in a different font lol
		local orig = FontType.PalatinoScaled
		FontType.PalatinoScaled = FontType.Default
		orig_updateHudPrints(self)
		FontType.PalatinoScaled = orig
	end
	--[[
	for i,ray in ipairs(debugDrawRays) do
		local endpos = ray[1]+ray[2]*ray[3]
		DebugDraw.drawLine3(ray[1], endpos, Color.Green)
		local size = vec(0.1,0.1,0.1)
		DebugDraw.drawBox3(endpos-size,endpos+size) 
	end
	debugDrawRays = {}
	]]
	
end
--[[
function addDebugDrawRay(origin, dir, dist)
	debugDrawRays[#debugDrawRays+1] = {origin, dir, dist}
end
]]

--------------------------------------------------------------------------------
--
-- clicking doesnt do much in this mod unless its the middle mouse button
--
--------------------------------------------------------------------------------
local zoomIn = false
function GameMode:mousePressed(button, x, y)
 	local viewport = self.viewport
 	if self.paused or x < 0 or x >= viewport[3] or y >= viewport[4] or y < 0 then return end
 	if infoMode then infoMode = false end
 	if button == 1 then zoomIn = not zoomIn end
end

function PartyComponent:reachableEntities()
	local map = self.go.map
	
	-- duplicates Map:allEntities() but with two return values as
	-- pickComponents() expects. yeah it's sad we can't just return
	-- allEntities()...
	local ents = map.entities
	local i = 1
	return function()
		while i <= map.maxEntities do
			local e = ents[i]
			i = i + 1
			if e then return i,e end
		end
		return nil
	end
end


local partyMovementVelocity = vec()

local lastKeyPressTime = {}

local linwfrKeyPos = vec(7.026474,0,6.508396)
local wallscrollBoxSize = vec(1.1,0.8,0.01)
local tinyTextBoxSize = vec(0.1,0.075,0.01)

function PartyComponent:update()
	local deltaTime = Time.deltaTime
--------------------------------------------------------------------------------
--
-- Add zoom-in
--
--------------------------------------------------------------------------------
	local targetFov = 1.1106
	local camera = self.camera
	local cameraRotation = tmat.create(camera:fetchVector("rotation_x",tvec()),
		camera:fetchVector("rotation_y",tvec()),
		camera:fetchVector("rotation_z",tvec()),
		tvec())
	if zoomIn then
		-- restrict zooming in 
		local cameraPos = camera:fetchVector("position",tvec())
		-- warning, this only works because the camera's parent nodes
		-- have no rotation/shear transform. if that can't be assumed
		-- then replace with getWorldRotation() (churns more memory)
		local cameraDirection = cameraRotation.z
		local wallscroll = findObject("wallscroll")
		if wallscroll then
			tempRay.pos = cameraPos
			tempRay.dir = cameraDirection
			tempBox.pos = wallscroll:getWorldPositionFast()
			tempBox.hsize = wallscrollBoxSize
			if isectRayBox(tempRay,tempBox) then
				tempBox.pos = wallscroll.t111s.node:getWorldPosition()
				tempBox.hsize = tinyTextBoxSize
				if isectRayBox(tempRay,tempBox) then
					targetFov = 0.05
					-- not at all certain that the player actually noticed it but
					-- like whatever they're not gonna find the hint system anyway
					zoomedInOnTinyText = true
				else
					targetFov = 0.2
				end
			else
				targetFov = 0.8
			end
		end
	end
	if realTargetFov then targetFov = realTargetFov end
	
	-- framerate dependent in stupid ways and bouncy, but looks nice in practice
	-- as low as 20fps and as high as 1000fps, and nothing else looks nice at 20fps :p
	local fovchange = deltaTime*12
	self.cameraFov = self.cameraFov*(1-fovchange)+targetFov*(fovchange)
	
--------------------------------------------------------------------------------
--
-- Mouse look is always active now (except when paused) and isn't horizontally clamped
--
--------------------------------------------------------------------------------

-- do not use mouse movement from the same frame the mouse mode was changed, it's likely garbage
	if currentMouseMode == "relative" and lastMouseModeChangeFrame ~= Time.currentFrame then
		local dx,dy = sys_mousePos()
		if config.invertX then dx = -dx end
		if config.invertY then dy = -dy end
		
		-- yay random hardcoded values for everything...
		local k = targetFov < 0.5 and 0.0005 or 0.002

		self.cameraHeading = self.cameraHeading - dx * k
		
		self.cameraPitch = self.cameraPitch - dy * k
		self.cameraPitch = math_clamp(self.cameraPitch, math_rad(-50), math_rad(50))
	end
	
--------------------------------------------------------------------------------
--
-- Butcher movement controls, because we need the whole alphabet for incantations,
-- and everyone has cursor keys.
--
--------------------------------------------------------------------------------

	local up = sys_keyDown("up")
	local down = sys_keyDown("down")
	local right = sys_keyDown("right")
	local left = sys_keyDown("left")
	local lastUpTime = lastKeyPressTime["up"] or -1
	local lastDownTime = lastKeyPressTime["down"] or -1
	local lastLeftTime = lastKeyPressTime["left"] or -1
	local lastRightTime = lastKeyPressTime["right"] or -1
	
	local accelerationVector = tvec()
	if up and lastUpTime > lastDownTime then
		accelerationVector[3] = 1
	elseif down then
		accelerationVector[3] = -1
	end
	if right and lastRightTime > lastLeftTime then
		accelerationVector[1] = 1
	elseif left then
		accelerationVector[1] = -1
	end
	
	-- 1/[time it takes to reach top speed]
	local ACCEL_TIME_INV = 4
	-- 1/[time it takes to go from top speed to stationary]
	local DECEL_TIME_INV = 4
	-- top speed (m/s)
	local SPEED = 3
	
	-- this movement behavior has substantial inaccuracies, but since the
	-- exact speed of your movement in this mod is totally inconsequential
	-- i'm not going to bother handling acceleration correctly
	if vec_lengthSquared(accelerationVector) > 0.001 then
		accelerationVector[1], accelerationVector[2], accelerationVector[3] = mat_transformVectorNum(
			accelerationVector[1], accelerationVector[2], accelerationVector[3], cameraRotation)
		accelerationVector[2] = 0 -- no vertical movement
		vec_normalizeInPlace3(accelerationVector)
		
		-- deceleration always occurs, so offset that here
		vec_mulInPlace(accelerationVector, ((ACCEL_TIME_INV+DECEL_TIME_INV) * deltaTime))
		vec_addInPlaceV(partyMovementVelocity, accelerationVector)
	end
	
	local newpos = self.go:getWorldPositionFast()
	local velocityLength = vec_length3(partyMovementVelocity)
	if velocityLength > 0.001 then
		local xsign = partyMovementVelocity[1]
		local zsign = partyMovementVelocity[3]
		local mult = 1 - (1 / velocityLength * (DECEL_TIME_INV * deltaTime))
		partyMovementVelocity[1] = partyMovementVelocity[1]*mult
		partyMovementVelocity[3] = partyMovementVelocity[3]*mult
		if xsign*partyMovementVelocity[1] < 0 then partyMovementVelocity[1] = 0 end
		if zsign*partyMovementVelocity[3] < 0 then partyMovementVelocity[3] = 0 end
		
		local distance = vec_length3(partyMovementVelocity)
		if distance > 0.001 then
			local dirx = partyMovementVelocity[1]/distance
			local dirz = partyMovementVelocity[3]/distance
			if distance > 1 then -- speed cap
				partyMovementVelocity[1] = dirx
				partyMovementVelocity[3] = dirz
				distance = 1
			end
			
			local mult = distance*deltaTime*SPEED
			
			newpos[1] = math_clamp(newpos[1]+dirx*mult,ROOM_WALL_MARGIN,ROOM_SIZE-ROOM_WALL_MARGIN)
			newpos[3] = math_clamp(newpos[3]+dirz*mult,ROOM_WALL_MARGIN,ROOM_SIZE-ROOM_WALL_MARGIN)
		end
	end
	
	-- quick "collision" hack so that you cant just walk inside the figurines
	-- this is a horrible movement model, don't do this. it assumes no overlapping
	-- spheres, doesn't conserve energy, and doesn't even make sure to push you
	-- out of the correct side of the sphere
	for _,c in self.go.map:allComponents(ProjectileColliderComponent) do
		local sphereOrigin = c.node:getWorldPosition()
		sphereOrigin[2] = newpos[2]
		local sphereRadius = c.size[1]
		local delta = sphereOrigin-newpos
		if vec_length3(delta) <= sphereRadius then
			-- push out to the edge of the sphere
			vec_normalizeInPlace3(delta)
			newpos = sphereOrigin-delta*sphereRadius
		end
	end
	
	-- HACK: magnetize the optimal position for viewing fej*sikomb's name
	if zoomIn then
		local linwfrKeyDist = vec_length3(newpos-linwfrKeyPos)
		if linwfrKeyDist < 0.2 then newpos = newpos*(0.8+linwfrKeyDist)+linwfrKeyPos*(0.2-linwfrKeyDist) end
	end
	self.go:setWorldPosition(newpos)
	
	self.cameraZOffset = 0
	
	if linwfrKeyEditMode then
		self.cameraFov = 0.05
	end
	self:updateCamera()
	
	if linwfrKeyEditMode then
		self.go:setWorldPosition(vec(7.026474,0,6.508396))
		self.camera:setRotation(mat_create(
			vec(0.793657, -0.000000, 0.608366, 0.000000),
			vec(-0.082370, 0.990792, 0.107458, 0.000000),
			vec(-0.602764, -0.135396, 0.786348, 0.000000),
			vec(0.000000, 0.000000, 0.000000, 1.000000)))
	end

	self.cameraPositionOffset.x = 0
        self.cameraPositionOffset.y = 0
        self.cameraPositionOffset.z = 0
        self.cameraRotationOffset.x = 0
        self.cameraRotationOffset.y = 0
        self.cameraRotationOffset.z = 0
end

--------------------------------------------------------------------------------
--
-- Remove fog and increase far clip a bit (we don't want anything culled by
-- distance here).
--
--------------------------------------------------------------------------------
function GameMode:updateFog()
	renderer:setFog(false)
end
function GameMode:getFarClip()
	return 1000
end
-- grimrock.cfg can set this to really low values, don't allow that here.
-- Don't worry, the game never *saves* maxDrawDistance to grimrock.cfg.
config.maxDrawDistance = 1000

--------------------------------------------------------------------------------
--
-- Stuff for specific incantations
--
--------------------------------------------------------------------------------

#define GODDESS_SUMMON_TIME 2
local invokedGoddesses = {}
local function invokeGoddess(goddess,dontSave)
	if invokedGoddesses[goddess] then return false end
	invokedGoddesses[goddess] = true
	if not dontSave then burdenOfKnowledgeSave() end
	
	local color = goddessData[goddess].color
	color = goddess == FINAL_NUMBER and {1,1,1} or {color[1],color[2],color[3]}
	
	-- change negative colors to their complement
	for i=1,3 do
		local col = color[i]
		if col < 0 then
			color[i] = 0
			color[i%3+1] = -col
			color[(i+1)%3+1] = -col
			break
		end
	end
	
	local particleSystemName = "invoke_"..goddess
	defineParticleSystem{
		name = particleSystemName,
		emitters = {
			{
				emitterShape = "MeshShape",
				maxParticles = 4096,
				emissionRate = 1024,
				sprayAngle = {0, 0},
				velocity = {0, 0},
				texture = "assets/textures/common/white.tga",
				lifetime = {64.0, 64.0},
				color0 = color,
				fadeIn = 0.2,
				fadeOut = 0.2,
				size = {0.004, 0.004},
				gravity = {0,0,0},
				airResistance = 1,
				rotationSpeed = 0,
				blendMode = "Additive",
				randomInitialRotation = false,
			},
		},
	}
	defineParticleSystem{
		name = particleSystemName.."_finish",
		emitters = {
			{
				emitterShape = "MeshShape",
				maxParticles = 1024,
				spawnBurst = true,
				sprayAngle = {0, 180},
				velocity = {0.3, 2},
				texture = "assets/textures/common/white.tga",
				lifetime = {1.0, 5.0},
				color0 = color,
				fadeIn = 0.001,
				fadeOut = 1.0,
				size = {0.004, 0.004},
				gravity = {0,-0.1,0},
				airResistance = 1,
				rotationSpeed = 0,
				blendMode = "Additive",
				randomInitialRotation = false,
			},
		},
	}
	
	local fx = spawn(dungeon.maps[1],"goddess_invoke_fx",0,0,0,0,"goddess_invoke_"..goddess)
	fx:setWorldPosition(goddessData[goddess].position)
	fx.light.goddess = goddess -- bleh
	fx.light:setColor(goddessData[goddess].color)
	fx.particle:setParticleSystem(particleSystemName)
	fx.particle:setEmitterMesh(goddessData[goddess].mesh)
	
	return true
end
local function spawnLifeDeathHalfFX()
	local fx = spawn(dungeon.maps[1],"invoke_life_death_half",0,0,0,0,"invoke_life_death_half")
	fx:setWorldPosition(goddessData[LIFE_NUMBER].position)
	fx.particle:setEmitterMesh(goddessData[LIFE_NUMBER].mesh)
end
local function invokeLifeDeath(goddess)
	if invokedGoddesses[goddess] then return end
	invokedGoddesses[goddess] = true
	if invokedGoddesses[LIFE_NUMBER] and invokedGoddesses[DEATH_NUMBER] then
		local half = findObject("invoke_life_death_half")
		half.particle:stop() -- will destroy itself
		
		invokedGoddesses[LIFE_NUMBER] = false -- :/
		invokeGoddess(LIFE_NUMBER,true)
	else
		spawnLifeDeathHalfFX()
		soundSystem:playSound2D("goddess_invoke_half")
	end
	burdenOfKnowledgeSave()
	return true
end
local function spawnGoddess(goddess)
	invokedGoddesses[goddess] = true -- for loading savegames
	
	local o = spawn(dungeon.maps[1],goddessData[goddess].id,0,0,0,0,goddessData[goddess].id)
	o:setWorldPosition(goddessData[goddess].position)
	o.spot.node:lookAt(vec(4.5,1.5,8.75)) -- ASS
end
defineParticleSystem{
	name = "invoke_life_death_half",
	emitters = {
		{
			emitterShape = "MeshShape",
			maxParticles = 512,
			emissionRate = 256,
			sprayAngle = {0, 0},
			velocity = {0, 0},
			texture = "assets/textures/common/white.tga",
			lifetime = {1.0, 1.0},
			color0 = {LIFE_AND_DEATH_COLOR[1],LIFE_AND_DEATH_COLOR[2],LIFE_AND_DEATH_COLOR[3]},
			fadeIn = 0.2,
			fadeOut = 0.8,
			size = {0.004, 0.004},
			gravity = {0,0,0},
			airResistance = 1,
			rotationSpeed = 0,
			blendMode = "Additive",
			randomInitialRotation = false,
		},
	},
}
local function destroyCeiling()
	if findObject("ceiling_destroyer") or not findObject("ceiling") then return end
	
	spawn(dungeon.maps[1],"ceiling_destroyer",1,1,0,0,"ceiling_destroyer")
	soundSystem:playSound2D("ceiling_destruction")
	
	return true
end
local function destroyCeilingLogic()
	local track = dungeon.ambientTracks.background
	soundSystem:playStream2(track.filename,0)
	soundSystem:fadeStreamVolume(track.volume, 4)
	findObject("ceiling"):destroy()
	findObject("stars").model:enable()
	for _,c in findObject("starchart"):componentIterator() do
		c:enable()
	end
end

local function beginTheEnd()
	ending = true
	local o = spawn(dungeon.maps[1],"ending_light",0,0,0,0,"ending_light")
	o:setWorldPosition(4.5,1.5,4.5)
	gameMode:fadeOut(vec(),0)
	gameMode:fadeIn(vec(),1)
	soundSystem:fadeStreamVolume(0,2)
	for e in dungeon.maps[1]:allEntities() do
		if e.arch.name == "floor"
		or e.arch.name == "ceiling"
		or e.arch.name == "pedestal"
		or e.arch.name == "starchart"
		or e.arch.name == "wallscroll"
		or e.arch.name == "backwalltext"
		or e.arch.name == "ceiling_destroyer"
		or e.arch.name == "invoke_life_death_half"
		or e.arch.name:find("goddess") then
			e:destroyDelayed()
		end
	end
	math_randomseed(2000)
	local look = vec(4.5,1.5,4.5)
	for i=1,1024 do
		local o = spawn(dungeon.maps[1],"ending_text_message",0,0,0,0)
		local pos = vec(math_random()-0.5,math_random()-0.5,math_random()-0.5)
		vec_normalizeInPlace3(pos)
		vec_mulInPlace(pos,16+math_random()*16)
		o:setWorldPosition(pos)
		o.node:lookAt(look)
	end
	return true
end

--------------------------------------------------------------------------------
--
-- General incantation code
--
--------------------------------------------------------------------------------

local INCANTATIONS = {
["PHYS_NAME_LOWER"] = {function() return invokeGoddess(PHYS_NUMBER) end,PHYS_COLOR},
["WATER_NAME_LOWER"] = {function() return invokeGoddess(WATER_NUMBER) end,WATER_COLOR},
["EARTH_NAME_LOWER"] = {function() return invokeGoddess(EARTH_NUMBER) end,EARTH_COLOR},
["FIRE_NAME_LOWER"] = {function() return invokeGoddess(FIRE_NUMBER) end,vec(1,0,1)},
["AIR_NAME_LOWER"] = {function() return invokeGoddess(AIR_NUMBER) end,vec(1,1,0)},
["LIFE_NAME_LOWER"] = {function() return invokeLifeDeath(LIFE_NUMBER) end,vec(1,0,0)},
["DEATH_NAME_LOWER"] = {function() return invokeLifeDeath(DEATH_NUMBER) end,vec(1,0,0)},
["FINAL_NAME_LOWER"] = {function() return invokeGoddess(FINAL_NUMBER) end,vec(1,1,1)},
["CEILING_DESTRUCTION_INCANTATION"] = {function() return destroyCeiling() end,vec(0,1,1)},
["MUTE_INCANTATION"] = {function()
	config.muteMusic = true
	config.muteSounds = true
	config:apply()
	return true
end,vec(0.5,0,1)},
["UNMUTE_INCANTATION"] = {function()
	config.muteMusic = false
	config.muteSounds = false
	config:apply()
	return true
end,vec(1,0,0.5)},
["FINAL_INCANTATION"] = {function() beginTheEnd() end,vec(1,1,1)},
["hint"] = {function() hint() return true end,vec(1,0.5,0)},
}
local MAX_INCANTATION_LENGTH = 0
for incantation,f in pairs(INCANTATIONS) do
	MAX_INCANTATION_LENGTH = math_max(#incantation, MAX_INCANTATION_LENGTH)
end

local currentIncantation = ""

local currentIncantationChartLines = {}

local function defineIncantationChartLineParticles(name, length)
	defineZhandulEmitter{
		name = name,
		emissionRate = length*128,
		sprayAngle = {0, 26},
		boxMin = {0, 0, 0},
		boxMax = {0, 0, length},
		velocity = {0, 0.3},
		lifetime = {1.0, 2.0},
		color0 = {1.0, 1.0, 1.0},
		fadeIn = 0.0,
		fadeOut = 0.5,
		gravity = {0,0.4,0},
		airResistance = 0.8,
		rotationSpeed = 0,
		blendMode = "Additive",
		objectSpace = true,
		shader = "particles_pixel",
	}
--[[
	defineParticleSystem{
		name = name,
		emitters = {
			{
				emitterShape = "BoxShape",
				maxParticles = length*256,
				emissionRate = length*128,
				sprayAngle = {0, 12},
				boxMin = {0, 0, 0},
				boxMax = {0, 0, length},
				velocity = {0, 0.3},
				texture = "assets/textures/common/white.tga",
				lifetime = {1.0, 2.0},
				color0 = {1.0, 1.0, 1.0},
				fadeIn = 0.0,
				fadeOut = 0.5,
				size = {0.004, 0.004},
				gravity = {0,0,-0.4},
				airResistance = 0.8,
				rotationSpeed = 0,
				blendMode = "Additive",
				randomInitialRotation = false,
				objectSpace = true,
			},
			{
				emitterShape = "BoxShape",
				maxParticles = length+1,
				emissionRate = math_ceil(length*0.125),
				sprayAngle = {0, 60},
				boxMin = {0, 0, 0},
				boxMax = {0, 0, length},
				velocity = {0, 0.3},
				texture = "assets/textures/common/white.tga",
				lifetime = {1.0, 8.0},
				color0 = {1.0, 1.0, 1.0},
				fadeIn = 0.0,
				fadeOut = 1.0,
				size = {0.004, 0.004},
				gravity = {0,0.04,-0.05},
				airResistance = 0.1,
				rotationSpeed = 0,
				blendMode = "Additive",
				randomInitialRotation = false,
				objectSpace = true,
			},
		},
	}]]
	
end
local function getIncantationChartLineFXName(line)
	return "icl"..line
end
local function getNextLineFXComponentName(chart, line)
	local lineID = line
	local i = 1
	while chart:getComponent(lineID..i) do
		i = i+1
	end
	return lineID..i
end
local function getExistingLineFXComponents(chart, line)
	local lineID = line
	local i = 1
	local rval = {}
	for _,c in chart:componentIterator() do
		if c:getName():sub(1,2) == line then
			rval[#rval+1] = c
		end
	end
	return rval
end
local function addIncantationChartLineFX(line, num)
	if findObject("ceiling") then return end
	local chart = findObject("starchart")
	
	-- ZhandulParticleComponent does not support start() so we make extra components when needed instead
	--if existingLine then
		--existingLine:start()
	--else
		local char1 = line:sub(1,1)
		local char2 = line:sub(2,2)
		--local particleComponent = chart:createComponent(ParticleComponent, line)
		local particleComponent = chart:createComponent(ZhandulParticleComponent, getNextLineFXComponentName(chart, line))

		local particleSystemName = getIncantationChartLineFXName(line)
		if char1 ~= char2 then
			local char1Component = chart:getComponent(char1)
			if not char1Component then console:warn(char1.." not found") return end
			local char1Offset = char1Component:getOffset()
			local char2Component = chart:getComponent(char2)
			if not char2Component then console:warn(char2.." not found") return end
			local char2Offset = char2Component:getOffset()
			local size = vec_length3(char2Offset-char1Offset)
			--if not dungeon.particleSystems[particleSystemName] then
			if not ZhandulParticleComponent.emitterDefined(particleSystemName) then
				defineIncantationChartLineParticles(particleSystemName, size)
			end
			particleComponent:setOffset(char1Offset)
			particleComponent.node:lookAt(char2Component.node:getWorldPosition())
			--particleComponent.node:rotateAbout(0,MATH_PI*0.5)
			--particleComponent:setParticleSystem(particleSystemName)
			particleComponent:setEmitter(particleSystemName)
			particleComponent:setDestroySelf(true)
		end
	--end
end
local function removeIncantationChartLineFX(line)
	local chart = findObject("starchart")
	local ps = getExistingLineFXComponents(chart,line)
	for _,c in ipairs(ps) do
		c:stop()
	end
end
local function successfulIncantationChartLineFX(line,color)
	if findObject("ceiling") then return end
	local chart = findObject("starchart")
	local ps = getExistingLineFXComponents(chart,line)
	for _,c in ipairs(ps) do
		c:stop()
		c:setParam("g_particleColor",color,4)
		-- the final line will have a time of 0 so it won't really do
		-- anything without this hack. stupid and choppy but it's
		-- important that the shape of the incantation is clear and my
		-- self-imposed deadline for "ready for my girlfriend to playtest"
		-- is today.
		if c.particleTime < 1 then c.particleTime = 1+math_random() end
	end
end


local function getLinesForIncantation(incantation)
	local rval = {}
	for i=2,incantation:len() do
		local line = incantation:sub(i-1,i)
		rval[line] = (rval[line] or 0) + 1
	end
	return rval
end
local function updateIncantationChart()
	local desiredLines = getLinesForIncantation(currentIncantation)

	for line,_ in pairs(currentIncantationChartLines) do
		if (desiredLines[line] or 0) == 0 then
			removeIncantationChartLineFX(line)
			currentIncantationChartLines[line] = nil
		end
	end
	for line,_ in pairs(desiredLines) do
		if (currentIncantationChartLines[line] or 0) < desiredLines[line] then
			addIncantationChartLineFX(line, desiredLines[line])
			currentIncantationChartLines[line] = desiredLines[line]
		end
	end
end
local function incantCharScreenEffect(char, deleted)
	local o = spawn(dungeon.maps[1],"incant_char_effect",0,0,0,0)
	if deleted then
		o.isDeleteEffect = true
		o.zhandulparticle:setTimeScale(-4)
		o.zhandulparticle:setTime(1)
	end
	o.incantChar = char
	o.model:setText({char})
end
local function cancelIncantation()
	currentIncantation = ""
	updateIncantationChart()
end
local function finishIncantation(incantation)
	local desiredLines = getLinesForIncantation(incantation)
	
	for line,_ in pairs(desiredLines) do
		successfulIncantationChartLineFX(line,INCANTATIONS[incantation][2])
	end
	currentIncantation = ""
	updateIncantationChart()
end
local charToNote = {
	-- 
	q=57,
	w=58,
	e=59,
	r=60,
	t=61,
	y=62,
	u=63,
	i=64,
	o=65,
	p=67,
	
	a=69,
	s=70,
	d=71,
	f=72,
	g=73,
	h=74,
	j=75,
	k=76,
	l=77,
	
	z=81,
	x=82,
	c=83,
	v=84,
	b=85,
	n=86,
	m=87,
	
	-- let's not emulate what happens when you hit 8 in a tracker though
	["*"]=90,
}
local function charPitch(char)
	local note = charToNote[char] or 69
	note = note-12
	return 2^((note-69)/12)
end

local function incantationDeleteKey(volume)
	local deletedChar = currentIncantation:sub(-1,-1)
	if deletedChar ~= "" then
		if #currentIncantation == 1 then
			currentIncantation = ""
		else
			currentIncantation = currentIncantation:sub(1,-2)
		end
		-- we don't need to check if the new incantation is one with an
		-- effect, that would have been caught when it was originally
		-- typed, since the player can only add and remove from the end.
		updateIncantationChart()
		
		if #soundSystem.sounds < 200 then -- prevents mac players from crashing the game by typing 256 characters in rapid succession
			soundSystem:playSound2D("keypress_delete",volume or 1,charPitch(deletedChar))
		end
		incantCharScreenEffect(deletedChar,true)
	end
end
-- returns true if the key event was an incantationy one (alphabetical or star)
local function handleIncantationKeyPress(event)
	if ending then return end
	local char = event.char and string.lower(event.char)
	local star = char == "*" or char == "8"
	if event.key == "enter" then
		local len = #currentIncantation
		for i=1,len do incantationDeleteKey(1/len) end
	elseif event.key == "backspace" or event.key == "delete" then
		incantationDeleteKey()
	elseif char and char >= "a" and char <= "z" or star then
		local incantChar = star and "*" or char
		currentIncantation = currentIncantation..incantChar
		if #currentIncantation > MAX_INCANTATION_LENGTH then
			currentIncantation = currentIncantation:sub(#currentIncantation-MAX_INCANTATION_LENGTH+1)
		end
		
		updateIncantationChart()
		
		for incantation,data in pairs(INCANTATIONS) do
			if currentIncantation:sub(-#incantation) == incantation and data[1]() then
				finishIncantation(incantation)
				break
			end
		end
		
		if #soundSystem.sounds < 200 then -- prevents mac players from crashing the game by typing 256 characters in rapid succession
			soundSystem:playSound2D("keypress",1,charPitch(incantChar))
		end
		incantCharScreenEffect(incantChar)
	else
		return false
	end
	--console:print(currentIncantation)
	return true
end

local function doGoddessFlip(num, instant)
	local goddess = findObject(goddessData[num].id)
	if goddess then
		if goddess.model:isEnabled() and not tween_isMoving(goddess) then
			goddess.flipped = not goddess.flipped
			
			local targetRotation = (goddess.flipped and MATH_PI or 0)
			tween_rotateOverTime(goddess, mat_rotationY(targetRotation), instant and 0.1 or 0.33, "SPHERICAL_SMOOTHERSTEP")
			if num ~= 9 and not instant then goddess:playSound("goddess_flip") end
		end
	end
end
local function handleGoddessFlip(event)
	if ending then return end
	local capitalNumbers = {
		["!"]=1,
		["@"]=2,
		["#"]=3,
		["$"]=4,
		["%"]=5,
		["^"]=6,
		["&"]=7,
		["*"]=8,
		["("]=9,
		[")"]=0,
	}
	-- DO NOT replace with event.keyCode, event.keyCode can be anything
	-- because of different keyboard locales
	local num = tonumber(event.char) or capitalNumbers[event.char]
	if num and goddessData[num] then
		doGoddessFlip(num)
	end
end

local currentLinwfrScale = 0.004
function GameMode:keyPressed(event)
	-- remove key-repeat behavior for keys that have names
	-- keys without names dont do anything barring really wack ass locales
	-- so let's not worry about the harder task of preventing it for those
	--
	-- BUG: yeah this wont really work if the player changes locales while
	-- holding down a key
	local name = sys.getKeyName(event.keyCode)
	if name then
		if keyStates[name] then return end
		keyStates[name] = true
	end
	
	local menu = self.menu
	if menu then
		if menu.keyPressed then return menu:keyPressed(event) end
		return true
	end
	
	if event.char == '+' or event.char == '=' then
		zoomIn = not zoomIn
	end
	
	if config.editor and event.key == "end" then
		linwfrKeyEditMode = not linwfrKeyEditMode
	end
	if linwfrKeyEditMode then
		if (event.char == "-" or event.char == "+") and linwfrKeyIndex then
			currentLinwfrScale = currentLinwfrScale+(event.char == "-" and -0.00003 or 0.00003)
			console:print(currentLinwfrScale)
			local o = findObject("linwfr_key_test_"..linwfrKeyIndex)
			o:setWorldRotationAngles(0,0,0)
			o.node:lookAt(party:getCamera():getWorldPosition())
			o:setWorldRotation(mat_mul(o:getWorldRotation(),mat_scaling(currentLinwfrScale,currentLinwfrScale,currentLinwfrScale)))
		elseif event.key == "enter" then
			linwfrKeyIndex = (linwfrKeyIndex or 0)+1
			local ray = getMouseRay()
			printTable(ray)
			local pos = ray.pos+ray.dir*(0.1+math_random()*0.4)
			print(pos)
			console:print(pos)
			local o = spawn(dungeon.maps[1],"linwfr_key_test",0,0,0,0,"linwfr_key_test_"..linwfrKeyIndex)
			o:setWorldPosition(pos)
			o.node:lookAt(party:getCamera():getWorldPosition())
			local scale = 0.0004
			o:setWorldRotation(mat_mul(o:getWorldRotation(),mat_scaling(scale,scale,scale)))
		end
	end
	if event.key == "escape" or (config.editor and event.key == "home") then
		infoMode = not infoMode
	end
	if gameDone and (event.char == "7" or event.char == "&" or event.key == "7") then
		endMessagePage = endMessagePage%#endMessagePages+1
	end
	
	handleGoddessFlip(event)
	
	-- we don't want other stuff interfering with incantations
	if not handleIncantationKeyPress(event) then
		local action = config:convertKeyToAction(event.keyCode)
		if action == "quick_save" then
			infoMessage(saveAttemptMessage)
		elseif action == "quick_load" then
			infoMessage(loadAttemptMessage)
		end
		if config.developer then self:debugKeyPressed(event) end
	end
	
	local ct = Time.currentTime
	lastKeyPressTime[event.key] = ct
end

--------------------------------------------------------------------------------
--
-- Shader setup
--
--------------------------------------------------------------------------------

local function defineShader(name, vertex, geometry, material, unlit, shadow)
	local shader = RenderableShader.create()
	shader:initLightPrePassRendererShader(vertex, geometry, material, unlit, shadow)
	ShaderEx.cache[name] = shader
end

-- FILESIZE: should be able to optimize this into the empty string, but requires
-- tweaking of blender export script (blender won't allow empty string as material name)
-- same for nodes in the model files - should we make the mesh the root node when possible?
local mainMaterial = "m"

local outlineMaterial = "o"

if config.renderEngine == "opengl" then
defineShader(mainMaterial,nil,nil,"mod_assets/m.fsh")
defineShader(outlineMaterial,"mod_assets/o.vsh",nil,"mod_assets/o.fsh","mod_assets/o.fsh")
defineShader("t",nil,nil,"mod_assets/t.fsh","mod_assets/t.fsh")
defineShader("conditional_text",nil,nil,"mod_assets/conditional_text.fsh","mod_assets/conditional_text.fsh")
defineShader("conditional_text_black",nil,nil,"mod_assets/conditional_text_black.fsh","mod_assets/conditional_text_black.fsh")
defineShader("unconditional_text",nil,nil,"mod_assets/unconditional_text.fsh","mod_assets/unconditional_text.fsh")
defineShader("c",nil,nil,nil,"mod_assets/c.fsh")
defineShader("s","mod_assets/s.vsh",nil,"mod_assets/s.fsh","mod_assets/s.fsh")
defineShader("hair","mod_assets/hair.vsh",nil,"mod_assets/hair.fsh","mod_assets/hair.fsh")
defineShader("bodyhair",nil,nil,nil,"mod_assets/bodyhair.fsh")
defineShader("eye",nil,nil,"mod_assets/eye.fsh","mod_assets/eye.fsh")
defineShader("light_sparkle","mod_assets/light_sparkle.vsh",nil,"mod_assets/light_sparkle.fsh","mod_assets/light_sparkle.fsh")
defineShader("particles_pixel","mod_assets/particles_pixel.vsh",nil,"mod_assets/particles_pixel.fsh","mod_assets/particles_pixel.fsh")
defineShader("particles_pixel_screen_burst","mod_assets/particles_pixel_screen_burst.vsh",nil,"mod_assets/particles_pixel.fsh","mod_assets/particles_pixel.fsh")
defineShader("particles_goddess_final","mod_assets/particles_goddess_final.vsh",nil,"mod_assets/particles_goddess_final.fsh","mod_assets/particles_goddess_final.fsh")
else
	-- yes this is like *basically* unreachable right now
	error("unsupported rendering engine: "..tostring(config.renderEngine).."\nplease restart with the opengl renderer")
end

defineMaterial{
	name = mainMaterial,
	doubleSided = false,
	lighting = true,
	alphaTest = false,
	ambientOcclusion = false,
	blendMode = "Opaque",
	shader = mainMaterial,
	textureAddressMode = "WrapU_ClampV_ClampW",
	glossiness = 0,
	depthBias = 0,
	ditherTex = "mod_assets/dither_patterns.tga",
}

defineMaterial{
	name = "bodyhair",
	diffuseMap = "assets/textures/common/white.tga",
	specularMap = "assets/textures/common/black.tga",
	doubleSided = false,
	lighting = false,
	alphaTest = false,
	ambientOcclusion = false,
	blendMode = "Translucent",
	shader = "bodyhair",
	textureAddressMode = "Wrap",
	glossiness = 0,
	depthBias = 0,
}

defineMaterial{
	name = outlineMaterial,
	doubleSided = false,
	lighting = false,
	alphaTest = false,
	castShadow = false,
	-- works better in the translucent render pass, in the opaque pass it
	-- tends to render behind things it shouldn't
	blendMode = "Additive",
	ambientOcclusion = false,
	shader = outlineMaterial,
}
-- pure white
defineMaterial{
	name = "w",
	diffuseMap = "assets/textures/common/white.tga",
	doubleSided = false,
	lighting = false,
	alphaTest = false,
	ambientOcclusion = false,
	blendMode = "Opaque",
	textureAddressMode = "Wrap",
	glossiness = 0,
	depthBias = 0,
}
defineMaterial{
	name = "h",
	doubleSided = false,
	lighting = false,
	alphaTest = false,
	ambientOcclusion = false,
	blendMode = "Opaque",
	textureAddressMode = "Wrap",
	glossiness = 0,
	depthBias = 0,
	shader = "hair",
}

-- pure black
defineMaterial{
	name = "b",
	diffuseMap = "assets/textures/common/black.tga",
	lighting = false,
	blendMode = "Opaque",
}

-- stars
-- expects a mesh full of 0x0 quads that will be expanded to 1 pixel, similar
-- yet different to the particle shader.
defineMaterial{
	name = "s",
	doubleSided = false,
	lighting = false,
	alphaTest = false,
	ambientOcclusion = false,
	blendMode = "Opaque",
	shader = "s",
}

-- text that only shows when lit appropriately
-- this is in the translucent pass instead of opaque with alpha testing because
-- there is always something opaque right behind it, so it can use lighting
-- without needing to actually go to the geometry buffer, and doesn't need
-- shadows, and being in the translucent pass means the shader doesn't need to
-- do discards. wow that was a mess of a sentence but idc enough to rewrite it
defineMaterial{
	name = "conditional_text",
	diffuseMap = "mod_assets/font.tga",
	doubleSided = false,
	lighting = false,
	alphaTest = false,
	ambientOcclusion = false,
	blendMode = "Translucent",
	textureAddressMode = "Clamp",
	glossiness = 0,
	depthBias = 0,
	alphaBias = 0,
	shader = "conditional_text",
}
defineMaterial{
	name = "conditional_text_tiny",
	diffuseMap = "mod_assets/font.tga",
	doubleSided = false,
	lighting = false,
	alphaTest = false,
	ambientOcclusion = false,
	blendMode = "Translucent",
	textureAddressMode = "Clamp",
	glossiness = 0,
	depthBias = 0,
	alphaBias = 0.3,
	shader = "conditional_text",
}
defineMaterial{
	name = "conditional_text_black",
	diffuseMap = "mod_assets/font.tga",
	doubleSided = false,
	lighting = false,
	alphaTest = false,
	ambientOcclusion = false,
	blendMode = "Translucent",
	textureAddressMode = "Clamp",
	glossiness = 0,
	depthBias = 0,
	shader = "conditional_text_black",
}
-- text that always shows
defineMaterial{
	name = "unconditional_text",
	diffuseMap = "mod_assets/font.tga",
	doubleSided = false,
	lighting = false,
	alphaTest = false,
	ambientOcclusion = false,
	blendMode = "Translucent",
	textureAddressMode = "Clamp",
	glossiness = 0,
	depthBias = 0,
	shader = "unconditional_text",
}

-- additive emissive color
defineMaterial{
	name = "c",
	diffuseMap = "assets/textures/common/black.tga",
	lighting = false,
	blendMode = "Additive",
	shader = "c",
}

defineMaterial{
	name = "eye",
	doubleSided = false,
	lighting = false,
	alphaTest = false,
	ambientOcclusion = false,
	blendMode = "Opaque",
	depthBias = 0,
	shader = "eye",
}

function MaterialEx.updateAllMaterials()
	local ct = Time.currentTime
	for _,mat in pairs(dungeon.materials) do
		if mat.onUpdate then
			mat.onUpdate(objectToProxy(mat), ct)
		end
		mat:setParam("time", ct)
	end
end

--------------------------------------------------------------------------------
--
-- Object definition setup
--
--------------------------------------------------------------------------------
-- functions assigned to fields starting with "on" in object definitions get
-- setfenved to globalHookEnv
globalHookEnv = _G

local outlineModels = {}

ModelComponent.synthesizeProperty("outline")
ModelComponent.synthesizeProperty("text")
-- this always gets called after setModel because of ModelComponent:loadProperties()
-- XXX: setMaterial() breaks this lol (it changes the material of the outline entity too)
function ModelComponent:setOutline(outline)
	if self.outline == outline then return end

	if outline then
		local filename = self.modelFilename
		if not outlineModels[filename] then
			-- this clone probably isn't necessary
			local mesh = Mesh.load(self.modelFilename):clone()
			-- flip normals
			mesh:flipFaces()
			mesh:computeVertexNormals()
		
			mesh:setMaterial(dungeon.materials[outlineMaterial].material)
			outlineModels[filename] = RenderableMesh.create(mesh)
		end
		local me = MeshEntity.create(outlineModels[filename])
		self.outlineNode = self.go.map.scene:addNode()
		self.outlineNode.RenderEntity = me
		self.outlineNode:addTo(self.node)
	else
		-- NYI: un-outlining is pretty pointless.
	end
	self.outline = outline
end

local CHARACTER_WIDTHS = {}
for c=32,126 do
	CHARACTER_WIDTHS[c] = 0.05
end

#define FONT_TEXTURE_SIZE 2048

-- super hacky, every property of the font texture is hardcoded also there's no
-- support for non-monospace (I knew I wanted monospace for the puzzles anyway!)
--
-- funny thing is, I'd already written a non-monospace-only thing like this for
-- Avenie and just decided not to use it here I guess?
local textMeshCache = {}
function ModelComponent:setText(text)
	if not self.material then error("ModelComponent text requires a 'material' field") end
	local material = dungeon.materials[self.material]
	if not material then error("ModelComponent text: material "..self.material.." does not exist") end
	material = material.material

	local vertexArray = {}
	local texcoordArray = {}
	local indices = {}
	local STARTX,STARTY,WIDTH,HEIGHT,MARGIN,XOFF,YOFF = 1,2,3,4,5,6,7
	local CHARS_PER_ROW = 15
	local CHAR_WIDTH_PIXELS = 136
	local CHAR_WIDTH_FRACTION = CHAR_WIDTH_PIXELS/FONT_TEXTURE_SIZE
	local CHAR_HEIGHT_PIXELS = 218
	local CHAR_HEIGHT_FRACTION = CHAR_HEIGHT_PIXELS/FONT_TEXTURE_SIZE
	local CHARACTER_WIDTH = 0.05
	local CHARACTER_SPACING = 0
	local LINE_HEIGHT = CHARACTER_WIDTH*2
	local scale = 1
	local offsetX = 0
	local offsetY = 0
	local centerX = true
	local centerY = true
	if type(text) == "table" then
		scale = text.scale or 1
		offsetX = text.offset and text.offset[1]*CHARACTER_WIDTHS[32] or 0
		offsetY = text.offset and text.offset[2]*LINE_HEIGHT or 0
		text = text[1]
	end
	local cacheEntry = textMeshCache[text]
	local renderableMesh
	if cacheEntry then
		for i,entry in ipairs(cacheEntry) do
			if entry.scale == scale and entry.offsetX == offsetX and entry.offsetY == offsetY then
				renderableMesh = entry.renderable
				break
			end
		end
	end

	if not renderableMesh then
		local lineArray = text:split("\n")
		
		local ypos = #lineArray*LINE_HEIGHT*(centerY and 0.5 or 1)*scale+offsetY
		local texcoordMargin = 2/FONT_TEXTURE_SIZE -- prevents bleeding from adjacent characters
		for li,line in ipairs(lineArray) do
			local length = line:len()
			local xpos = offsetX
			if centerX then xpos = xpos-length*0.5*CHARACTER_WIDTHS[32]*scale end
			for i=1,length do
				local c = line:byte(i)
				local width = CHARACTER_WIDTHS[c]
				if not width then error("invalid character: "..string.char(c).." ("..c..")".." in text "..text) end
				width = width*scale
				
				if c == 32 then
					-- space
					xpos = xpos+width
				else			
					-- Left half of spacing
					if i ~= 1 then xpos = xpos+CHARACTER_SPACING*scale end
					
					-- Make a rectangle for the character.
					
					-- vertices
					local nwi = #vertexArray+1
					local nei = nwi+3
					local swi = nwi+6
					local sei = nwi+9
					local yhigh = ypos
					local ylow = yhigh-LINE_HEIGHT*scale
					vertexArray[nwi]=xpos
					vertexArray[nwi+1]=yhigh
					vertexArray[nwi+2]=0
					vertexArray[swi]=xpos
					vertexArray[swi+1]=ylow
					vertexArray[swi+2]=0
					xpos = xpos+width
					vertexArray[nei]=xpos
					vertexArray[nei+1]=yhigh
					vertexArray[nei+2]=0
					vertexArray[sei]=xpos
					vertexArray[sei+1]=ylow
					vertexArray[sei+2]=0
					
					-- texcoords
					local ti = #texcoordArray+1
					local fontIndex = c-33
					local fontIndexX = fontIndex%CHARS_PER_ROW
					local fontIndexY = math_floor(fontIndex/CHARS_PER_ROW)
					local westx = CHAR_WIDTH_FRACTION*fontIndexX-texcoordMargin
					local eastx = westx+CHAR_WIDTH_FRACTION+texcoordMargin
					local northy = fontIndexY*CHAR_HEIGHT_FRACTION+texcoordMargin
					local southy = northy+CHAR_HEIGHT_FRACTION-texcoordMargin
					texcoordArray[ti]=westx -- NW
					texcoordArray[ti+1]=northy
					texcoordArray[ti+2]=eastx -- NE
					texcoordArray[ti+3]=northy
					texcoordArray[ti+4]=westx -- SW
					texcoordArray[ti+5]=southy
					texcoordArray[ti+6]=eastx -- SE
					texcoordArray[ti+7]=southy
					
					-- Connect the vertices with two triangles
					local ii = #indices+1
					local sii = (nwi-1)/3 -- NW vertex
					indices[ii] = sii -- NW-NE-SW triangle
					indices[ii+1] = sii+1
					indices[ii+2] = sii+2
					indices[ii+3] = sii+3 -- NE-SW-SE triangle
					indices[ii+4] = sii+2
					indices[ii+5] = sii+1
					
					-- Right half of spacing
					if i ~= length then xpos = xpos+CHARACTER_SPACING*scale end
				end
			end
			ypos = ypos-LINE_HEIGHT*scale
		end
		local textMesh = Mesh.create(#vertexArray/3)
		textMesh:setVertexArray(vertexArray)
		textMesh:setTexcoordArray(texcoordArray)
		textMesh:setIndices(indices)
		textMesh:addSegment(material,"triangles",0,#indices/3)
		textMesh:computeVertexNormals()
		textMesh:computeTangentVectors() -- Needed for normal mapping
		renderableMesh = RenderableMesh.create(textMesh)
		local entry = {scale=scale,offsetX=offsetX,offsetY=offsetY,renderable=renderableMesh}
		if cacheEntry then
			cacheEntry[#cacheEntry+1] = entry
		else
			textMeshCache[text] = {entry}
		end
	end
		
	local me = MeshEntity.create(renderableMesh)
	me:setMaterial(material)
	self.node.RenderEntity = me
	if self.emissiveColor then me:setEmissiveColor(self.emissiveColor) end
	self.node:addTo(self.go.node)
	if not self.enabled then hideNode(self.node, true) end
end
-- FILESIZE: we probably don't need storeSourceData or lods, and can optimize
-- this emissive color hack a lot
function ModelComponent:loadProperties(desc)
	self.storeSourceData = desc.storeSourceData
	if desc.model then self:setModel(desc.model) end
	if desc.lodModel then self:setLodModel(desc.lodModel) end
	if desc.emissiveColor then self:setEmissiveColor(desc.emissiveColor) end
	if desc.material then self:setMaterial(desc.material) end
	Component.loadProperties(self, desc)
end

local orig_defineObject = defineObject
local function defineObject(def)
	orig_defineObject(def)
end

local cumulativeFaceAreaCache = {}
-- returns {x1,y1,z1,x2,y2,z2,...}
-- XXX: this is horrifically slow
local function getRandomPointsOnMesh(mesh, name, count, rejectByDistanceStart, rejectByDistanceEnd)
	local vertexArray = mesh:getVertexArray()
	local indices = mesh:getIndices()
	local cumulativeFaceAreas = name and cumulativeFaceAreaCache[name]
	if not cumulativeFaceAreas then
		cumulativeFaceAreas = {}
		local totalArea = 0.0
		for i=1,#indices/3 do
			local vert0i = indices[i*3-2]
			local vert0 = vec(vertexArray[vert0i*3+1],vertexArray[vert0i*3+2],vertexArray[vert0i*3+3])
			local vert1i = indices[i*3-1]
			local vert1 = vec(vertexArray[vert1i*3+1],vertexArray[vert1i*3+2],vertexArray[vert1i*3+3])
			local vert2i = indices[i*3]
			local vert2 = vec(vertexArray[vert2i*3+1],vertexArray[vert2i*3+2],vertexArray[vert2i*3+3])
			cumulativeFaceAreas[i] = totalArea
			totalArea = totalArea + vec_length3(vec_cross(vert0-vert1,vert0-vert2))
		end
		-- final element is the total area and doesn't correspond to a face
		cumulativeFaceAreas[#cumulativeFaceAreas+1] = totalArea
		if name then cumulativeFaceAreaCache[name] = cumulativeFaceAreas end
	end
	
	local rval = {}
	local totalArea = cumulativeFaceAreas[#cumulativeFaceAreas]
	local faceCount = #cumulativeFaceAreas-1
	local i = 1
	while i <= count do
		local rand = math_random()*totalArea
		-- binary search to get the face
		local left = 0
		local right = faceCount-1
		local index = math_floor(faceCount/2)
		while left < right do
			local area = cumulativeFaceAreas[index+1]
			if area > rand then
				right = index-1
			else
				left = index
			end
			index = math_floor((left+right+1)/2)
		end
		
		-- get random point on face
		local vert0i = indices[index*3+1]
		local vert0 = vec(vertexArray[vert0i*3+1],vertexArray[vert0i*3+2],vertexArray[vert0i*3+3])
		local vert1i = indices[index*3+2]
		local vert1 = vec(vertexArray[vert1i*3+1],vertexArray[vert1i*3+2],vertexArray[vert1i*3+3])
		local vert2i = indices[index*3+3]
		local vert2 = vec(vertexArray[vert2i*3+1],vertexArray[vert2i*3+2],vertexArray[vert2i*3+3])
		
		local a = math_random()
		local b = math_random()
		if a > b then
			local tmp = a
			a = b
			b = tmp
		end
		local point = vert0*a+vert1*(b-a)+vert2*(1-b)
		
		-- CATASTROPHIC HACK: this bit of weapons-grade ooftonium exists as an easy way
		-- to make Linwfr's name look better, at the cost of being the most disgusting
		-- thing I've ever written (not counting forum posts)
		local reject
		if rejectByDistanceStart then
			local len = vec_length3(point)
			local threshold = rejectByDistanceStart+(math_random()^6)*(rejectByDistanceEnd-rejectByDistanceStart)
			if len > threshold then
				reject = true
			end
		end

		if not reject then
			rval[i*3-2] = point[1]
			rval[i*3-1] = point[2]
			rval[i*3] = point[3]
			i = i+1
		end
	end
	return rval
end


--------------------------------------------------------------------------------
--
-- Object definitions
--
--------------------------------------------------------------------------------
defineZhandulEmitter{
	name = "linwfr_key_test",
	emissionRate = 64,
	sprayAngle = {0, 360},
	boxMin = {0, 0, 0},
	boxMax = {0, 0, 0},
	velocity = {0.2, 0.2},
	lifetime = {1.0, 1.0},
	color0 = {1.0, 1.0, 1.0},
	fadeIn = 0.0,
	fadeOut = 0.5,
	gravity = {0,0,0},
	airResistance = 1000,
	rotationSpeed = 0,
	blendMode = "Additive",
	objectSpace = true,
	shader = "particles_pixel",
}

defineObject{
	name = "linwfr_key_test",
	components = {
		{
			class = "Model",
			model = "mod_assets/linwfr_key_circle.fbx",
			material = "eye",
			emissiveColor = EARTH_COLOR,
		},
	},
	placement = "floor",
}



defineObject{
	name = "ending_light",
	components = {
		{
			class = "Light",
			color = vec(-15,-15,-15),
			range = 128,
			brightness = 0,
			onUpdate = function(self)
				self = proxyToObject(self)
				local lastTime = self.time or 0
				local time = lastTime+Time.deltaTime
				self.time = time
				
				self:setBrightness(math_min(time*0.2,1))
				dungeon.materials.s.material:setParam("g_endTime",time*14000)
				
				if lastTime <= 13 and time > 13 then
					gameMode:fadeOut(vec(),2)
				elseif lastTime <= 15 and time > 15 then
					for e in dungeon.maps[1]:allEntities() do
						if e.arch.name == "roomwall"
						or e.arch.name == "ending_text_message"
						or e.arch.name == "stars" then
							e:destroy()
						end
					end
					gameMode:fadeIn(vec(),1)
					gameDone = true
					self.go:destroy()
				end
			end,
		},
	},
	placement = "floor",
}
defineObject{
	name = "ending_text_message",
	components = {
		{
			class = "Model",
			emissiveColor = vec(-1,-1,-1),
			text = {"and we're as good as dead.",scale=6},
			material = "unconditional_text",
			rotation = vec(180,0,180),
		},
	},
	placement = "floor",
}
local ceilingComponents = {
	{
		class = "Model",
		model = "mod_assets/ceiling.fbx",
	},
}
for x=0,2 do for y=0,2 do
ceilingComponents[#ceilingComponents+1] = {
	class = "Model",
	name = "t"..x..y,
	text = "CEILING_DESTRUCTION_INCANTATION",
	rotation = vec(90,180,0),
	offset = vec(x*3-3,4.2,y*3-3),
	material = "unconditional_text",
	emissiveColor = vec(1,1,1),
}
end end
defineObject{
	name = "floor",
	components = {
		{
			class = "Model",
			model = "mod_assets/floor.fbx",
		},
	},
	placement = "floor",
	
}
defineObject{
	name = "ceiling",
	components = ceilingComponents,
	placement = "floor",
}
defineObject{
	name = "ceiling_destroyer",
	components = {
		{
			class = "Timer",
			timerInterval = 2.95,
			onActivate = function(self)
				destroyCeilingLogic()
				burdenOfKnowledgeSave()
				self.go.particle:setParticleSystem("ceiling_destroyed") -- will destroy object
				self:disable()

				party:shakeCamera(0.03,1)
			end,
		},
		{
			class = "Timer",
			name = "t2",
			timerInterval = 1,
			onActivate = function(self)
				party:shakeCamera(0.02,2,2)
				self:disable()
			end,
		},
		{
			class = "Particle",
			particleSystem = "ceiling_destroying",
			emitterMesh = "mod_assets/ceiling.fbx",
			destroyObject = true,
		},
	},
	placement = "floor",
}
-- HACK: opengl renderer has a quite low maximum on the number of particles per
-- emitter, so use several copies of the same emitter for more particles...
local ceilingDestroyingEmitter = {
			emitterShape = "MeshShape",
			maxParticles = 4096,
			emissionRate = 1360,
			sprayAngle = {0, 0},
			velocity = {0, 0},
			texture = "assets/textures/common/white.tga",
			lifetime = {64.0, 64.0},
			color0 = {1,1,1},
			fadeIn = 0.2,
			fadeOut = 0.2,
			size = {0.01, 0.01},
			gravity = {0,0,0},
			airResistance = 1,
			rotationSpeed = 0,
			blendMode = "Additive",
			randomInitialRotation = false,
			depthBias = 0.02,
		}
local ceilingDestroyedEmitter = {
	maxParticles = 4096,
	spawnBurst = true,
	boxMin = {-4.5,3,-4.5},
	boxMax = {4.5,3.5,4.5},
	sprayAngle = {0, 180},
	velocity = {0.2, 2.6},
	texture = "assets/textures/common/white.tga",
	lifetime = {1.0, 16.0},
	color0 = {1,1,1},
	fadeIn = 0.001,
	fadeOut = 0.001,
	size = {0.01, 0.01},
	gravity = {0,-0.2,0},
	airResistance = 0.1,
	rotationSpeed = 0,
	blendMode = "Additive",
	randomInitialRotation = false,
}
defineParticleSystem{
	name = "ceiling_destroying",
	emitters = {
		ceilingDestroyingEmitter,
		ceilingDestroyingEmitter,
		ceilingDestroyingEmitter,
		ceilingDestroyingEmitter,
	},
}
defineParticleSystem{
	name = "ceiling_destroyed",
	emitters = {
		ceilingDestroyedEmitter,
		ceilingDestroyedEmitter,
		ceilingDestroyedEmitter,
		ceilingDestroyedEmitter,
		ceilingDestroyedEmitter,
		ceilingDestroyedEmitter,
		ceilingDestroyedEmitter,
		ceilingDestroyedEmitter,
	},
}

defineObject{
	name = "pedestal",
	components = {
		{
			class = "Model",
			model = "mod_assets/pedestal.fbx",
			outline = true,
		},
		{
			class = "ProjectileCollider",
			offset = vec(0,0,0),
			size = vec(0.5,0.5,0.5),
		},
	},
	placement = "floor",
}

defineZhandulEmitter{
	name = "incant_char_screen",
	maxParticles = 256,
	sprayAngle = {0, 360},
	boxMin = {0, 0, 0},
	boxMax = {0, 0, 0},
	velocity = {0.025, 0.025},
	lifetime = {0.5, 1.0},
	color0 = {1.0, 1.0, 1.0},
	fadeIn = 0.0,
	fadeOut = 0.5,
	gravity = {0,0,0},
	airResistance = 2.6,
	rotationSpeed = 0,
	blendMode = "Additive",
	objectSpace = true,
	shader = "particles_pixel_screen_burst",
}
defineObject{
	name = "incant_char_effect",
	components = {
		{
			class = "Model",
			emissiveColor = vec(1,1,1),
			material = "unconditional_text",
			-- incant code sets text
		},
		{
			class = "ZhandulParticle",
			emitter = "incant_char_screen",
		},
		{
			class = "Timer",
			timerInterval = 0,
			onInit = function(self)
				-- as fractions of width/height instead of
				-- pixels so that we're resolution independent
				self = proxyToObject(self)
				self.screenX = 0.1+math_random()*0.8
				self.screenY = 0.1+math_random()*0.8
			end,
			onActivate = function(self)
				self = proxyToObject(self)
				self.time = (self.time or 0) + Time.deltaTime
				local dur = self.go.isDeleteEffect and 0.3 or 0.7
				local normalizedTime = self.time/dur
				if normalizedTime > 1 then
					self.go.model:disable()
					if self.go.zhandulparticle:isDone() then self.go:destroy() return end
				end
				
				local rotation = mat_rotationZ(self.go.isDeleteEffect and MATH_PI or 0)
				local scale = (1-normalizedTime)*0.125
				local starchart = findObject("starchart")
				if not starchart then self.go:destroy() return end
				local chartComponent = starchart:getComponent(self.go.incantChar)
				if chartComponent and chartComponent.enabled then
					local pos = chartComponent.node:getWorldPosition()
					pos[2] = pos[2]+0.001
					self.go:setWorldPosition(pos)
					rotation = mat_mul(rotation, mat_scaling(48,48,48))
					rotation = mat_mul(rotation, mat_rotationX(MATH_PI*1.5))
				else
					local ray = party:getCamera():getWorldRay(vec(self.screenX, self.screenY))
					-- if i got much closer than this my badly chosen near clip would bite me in the ass, which is a privilege reserved for my girlfriend
					self.go:setWorldPosition(ray.pos+ray.dir*0.125)

					rotation = mat_mul(rotation,party:getCamera():getWorldRotation())
				end

				self.go.model.node:setRotation(mat_mul(rotation,mat_scaling(scale,scale,scale)))
				self.go.zhandulparticle.node:setRotation(rotation)
			end,
		},
	},
	placement = "floor",
}

local function goddessBaseComponent(num)
	return {
			class = "Model",
			name = "b",
			model = "mod_assets/b"..num..".fbx",
			outline = true,
		}
end
local function goddessSoundComponent(num)
	return {
		class = "Sound",
		sound = "goddess_noise_"..num,
		-- not that y offset really matters in grimrock's sound system...
		offset = vec(0,APPROX_GODDESS_HEIGHT*0.6,0),
	}
end
local function goddessBaseLabelComponent(text)
	return {
			class = "Model",
			name = "bl",
			text = {text,scale=0.3},
			material = "unconditional_text",
			emissiveColor = vec(1,1,1),
			offset = vec(0,0.013,0),
			rotation = vec(270,0,0),
		}
end
local lightCollisionBox = Mesh.createBox(8.5,8.5,8.5)
lightCollisionBox:flipFaces()
-- blender-to-grimrock space for offsets: -Y, Z, X
local function goddessSpotLightComponent(offset, color, orbitCount, orbitCenter, orbitTransform, orbitSpeed, orbitLightRange)
	local materialName = "light_sparkle_"..orbitCount
	if orbitCount > 0 then
		if not dungeon.materials[materialName] then
			defineMaterial{
				name = materialName,
				doubleSided = false,
				lighting = false,
				alphaTest = false,
				ambientOcclusion = false,
				blendMode = "Additive",
				glossiness = 0,
				depthBias = 0,
				shader = "light_sparkle",
				-- HACK: pure blue has poor visibility so make it bigger
				-- so that it's less poorly visible
				sparkleSize = orbitCount == WATER_NUMBER and 0.06 or 0.04,
				streakCount = orbitCount,
				streakWidth = orbitCount == WATER_NUMBER and 0.8 or 0.2,
				speed = 8/orbitCount,
			}
		end
	end

	-- I do not use an actual spotlight here for compatibility reasons; wine
	-- direct3D historically hasn't handled Grimrock's spotlights well.
	-- Instead, I raycast from the given offset to a slightly smaller version
	-- of the room walls and stick a point light there.
	--
	-- This also has the side effect of making the conditional text shaders
	-- much simpler, as the color will be uniform across the entire lit area
	-- when no goddesses are rotating (and when they're rotating the lights
	-- move pretty fast so we don't really mind if it looks weird for a bit).
	local orbitStarNodes = {}
	return {
			class = "Light",
			name = "spot",
			brightness = 80,
			range = 3,
			offset = offset,
			color = color,
			onInit = function(self)
				self = proxyToObject(self)
				for i=1,orbitCount do
					local l = self.go:createComponent("Light","ol"..i)
					l:setColor(color)
					l:setBrightness(3)
					l:setRange(orbitLightRange or 1.5)
					local p = self.go:createComponent("Model","om"..i)
					p:setModel(POINT_MODEL_FILE)
					p:setBoundBox({}) -- default custom bound box size/pos is fine
					p:setEmissiveColor(color)
					p:setMaterial(materialName)
					orbitStarNodes[i*2-1] = l.node
					orbitStarNodes[i*2] = p.node
				end
			end,
			onUpdate = function(self)
				Profiler.beginBlock("spotRaycast")
				self = proxyToObject(self)
				-- this position/direction determination is very inefficient...
				local go = self.go
				local gonode = go.node
				local origin = gonode:fetchVector("position",tvec())
				local tempMat = tmat.create(gonode:fetchVector("rotation_x",tvec()),
					gonode:fetchVector("rotation_y",tvec()),
					gonode:fetchVector("rotation_z",tvec()),
					tvec())
				local offx,offy,offz = mat_transformVectorNum(offset[1],offset[2],offset[3],tempMat)
				origin[1] = origin[1]+offx-4.5
				origin[2] = origin[2]+offy
				origin[3] = origin[3]+offz-4.5
				local selfnode = self.node
				local dir = mat_toDirectionVector(selfnode:getWorldRotation())
				tempRay.pos = origin
				tempRay.dir = dir
				
				-- this is also very inefficient, it's a box, it doesn't need full mesh raycasting code...
				local hit,dist,index,normal = lightCollisionBox:raycast(tempRay, false)
				if not hit then
					dist = 1000
				end
				--addDebugDrawRay(origin,dir,dist)
				
				selfnode:fetchVector("rotation_x",tempMat.x)
				selfnode:fetchVector("rotation_y",tempMat.y)
				selfnode:fetchVector("rotation_z",tempMat.z)
				-- the offset from the 'offset' local variable...god this is pants
				local x,y,z = mat_transformVectorNum(0,0,dist,tempMat)
				selfnode:setPosition(tvec(offset[1]+x,offset[2]+y,offset[3]+z))
				Profiler.endBlock("spotRaycast")
			
				Profiler.beginBlock("spotOrbits")
				if orbitCount > 0 then
					local ocx, ocy, ocz = orbitCenter[1], orbitCenter[2], orbitCenter[3]
					local time = Time.currentTime*orbitSpeed
					for i=1,orbitCount do
						local t = (time+i/orbitCount)%1*(MATH_PI*2)
						local ox, oy, oz = mat_transformVectorNum(math_sin(t), 0, math_cos(t), orbitTransform)
						local off = vec(ocx+ox,ocy+oy,ocz+oz)
						
						orbitStarNodes[i*2-1]:setPosition(off)
						orbitStarNodes[i*2]:setPosition(off)
					end
				end
				Profiler.endBlock("spotOrbits")
			end
		}
end
local function goddessEyeComponent(offset, color, side, num)
	local materialName = "eye_sparkle_"..num
	if not dungeon.materials[materialName] then
		defineMaterial{
			name = materialName,
			doubleSided = false,
			lighting = false,
			alphaTest = false,
			ambientOcclusion = false,
			blendMode = "Additive",
			glossiness = 0,
			depthBias = 0,
			shader = "light_sparkle",
			sparkleSize = 0.02,
			streakCount = num,
			streakWidth = 0.6,
		}
	end
	return {
		class = "Model",
		name = "eyep"..side,
		offset = offset,
		model = POINT_MODEL_FILE,
		emissiveColor = color,
		material = materialName,
	}
end
defineObject{
	name = "goddess_invoke_fx",
	components = {
		{
			class = "Particle",
		},
		{
			class = "Sound",
			sound = "goddess_invoke",
			soundType = "ambient",
		},
		{
			class = "Light",
			range = 2,
			brightness = 10,
			offset = vec(0,0.5,0),
			onUpdate = function(self)
				self = proxyToObject(self)
				local lastTime = self.time or 0
				local time = lastTime+Time.deltaTime
				
				if lastTime < GODDESS_SUMMON_TIME and time >= GODDESS_SUMMON_TIME then
					spawnGoddess(self.goddess)
					local p = self.go.particle
					p:setParticleSystem(p:getParticleSystem().."_finish")
					p:setDestroyObject(true)
					self:setBrightness(20)
					self:setRange(3)
					self:setFadeOut(1)
					self:setDisableSelf(true)
				end
				
				self.time = time
			end,
		},
	},
	placement = "floor",
}
defineObject{
	name = "invoke_life_death_half",
	components = {
		{
			class = "Particle",
			particleSystem = "invoke_life_death_half",
			destroyObject = true, -- will only destroy after being stop()ed
		},
	},
	placement = "floor",
}
defineObject{
	name = "goddess_physicality",
	components = {
		{
			class = "Model",
			model = "mod_assets/gp.fbx",
			outline = true,
			emissiveColor = PHYS_COLOR,
		},
		{
			class = "Model",
			name = "h",
			model = "mod_assets/gph.fbx",
			emissiveColor = PHYS_COLOR,
		},
		goddessBaseComponent(PHYS_NUMBER),
		goddessSoundComponent(PHYS_NUMBER),
		goddessBaseLabelComponent([[


PHYS_NAME_LOWER
vii]]),
		--vec(0.0021,0.7534,0.0402)
		--vec(0.0259,0.7577,0.0318)
		goddessEyeComponent(vec(-0.0021,0.7534,0.0402),PHYS_COLOR,0,PHYS_NUMBER),
		goddessEyeComponent(vec(-0.0259,0.7577,0.0318),PHYS_COLOR,1,PHYS_NUMBER),
		goddessSpotLightComponent(vec(-0.057,0.75,0.05),
			PHYS_COLOR,
			PHYS_NUMBER,
			vec(-0.014,0.717,0.007),
			mat_mul(mat_scaling(0.285,0.285,0.285),
				mat_rotationAbout(vec(-0.049,-0.111,-0.993),0.0933)),
			0.04)
	},
	placement = "floor",
}
defineObject{
	name = "goddess_water",
	components = {
		{
			class = "Model",
			model = "mod_assets/gw.fbx",
			outline = true,
			emissiveColor = WATER_COLOR,
		},
		{
			class = "Model",
			name = "h",
			model = "mod_assets/gwh.fbx",
			emissiveColor = WATER_COLOR,
		},
		goddessBaseComponent(WATER_NUMBER),
		goddessSoundComponent(WATER_NUMBER),
		goddessBaseLabelComponent([[
WATER_NAME_LOWER
iv]]),
		goddessEyeComponent(vec(0.063,0.717,0.154),WATER_COLOR,0,WATER_NUMBER),
		goddessSpotLightComponent(vec(-0.057,0.75,0.05),
			WATER_COLOR,
			WATER_NUMBER,
			vec(0.0897,0.5996,0.1910),
			mat_mul(mat_scaling(0.046,0.115,0.273),
				mat_rotationAbout(vec(-0.588,-0.681,-0.436),3.105)),
			0.3,
			0.8)
		--[[
		{
			class = "Light",
			brightness = 10,
			range = 3,
			offset = vec(-0.24,0.7,0.04),
			color = WATER_COLOR,
		},]]
	},
	placement = "floor",
}
local linwfrDiamondSize = 0.01
defineZhandulEmitter{
	name = "linwfr_diamond",
	emissionRate = 1024,
	sprayAngle = {0, 360},
	boxMin = {-linwfrDiamondSize, -linwfrDiamondSize, -linwfrDiamondSize},
	boxMax = {linwfrDiamondSize, linwfrDiamondSize, linwfrDiamondSize},
	velocity = {0, 0},
	lifetime = {1.0, 1.0},
	color0 = {0.0, 1.0, 0.0},
	fadeIn = 0.0,
	fadeOut = 1.0,
	gravity = {0,0,0},
	airResistance = 1,
	rotationSpeed = 0,
	blendMode = "Additive",
	objectSpace = true,
	shader = "particles_pixel",
}
defineZhandulEmitter{
	name = "linwfr_diamond_2",
	emissionRate = 256,
	sprayAngle = {89, 91},
	boxMin = {0,0,0},
	boxMax = {0,0,0},
	velocity = {24, 24},
	lifetime = {1.0, 1.0},
	color0 = {0.0, 1.0, 0.0},
	fadeIn = 0.0,
	fadeOut = 1.0,
	gravity = {0,0,0},
	airResistance = 1024,
	rotationSpeed = 0,
	blendMode = "Additive",
	objectSpace = true,
	shader = "particles_pixel",
}
local linwfrKeyPositions = {
	vec(6.958399,1.385514,6.595958), 0.00007,
	vec(6.770318, 1.345126, 6.839738), 0.00022,
	vec(6.940153, 1.381291, 6.621416), 0.0001,
	vec(6.936757, 1.380758, 6.623429), 0.0001,
	vec(6.909347, 1.374487, 6.660887), 0.0001,
	vec(6.924297, 1.377632, 6.642149), 0.0001,
	vec(6.804508, 1.350845, 6.802515), 0.00025,
	vec(6.959889, 1.385511, 6.594141), 0.00005,
	vec(6.815259, 1.353958, 6.780996), 0.00016,
	vec(6.871671, 1.366168, 6.708828), 0.00014,
}
local linwfrDiamondCenter = vec(0,0.535,0)
local linwfrComponents = {
	{
		class = "Model",
		model = "mod_assets/ge.fbx",
		outline = true,
		emissiveColor = EARTH_COLOR,
	},
	{
		class = "Model",
		name = "h",
		model = "mod_assets/geh.fbx",
		emissiveColor = EARTH_COLOR,
	},
	goddessBaseComponent(EARTH_NUMBER),
	goddessSoundComponent(EARTH_NUMBER),
	goddessBaseLabelComponent([[
EARTH_NAME_LOWER
ii]]),
	goddessEyeComponent(vec(0.12125,0.63881,0.06421),EARTH_COLOR,0,EARTH_NUMBER),
	goddessEyeComponent(vec(0.13299,0.63201,0.08987),EARTH_COLOR,1,EARTH_NUMBER),
	goddessSpotLightComponent(vec(0.11591,0.64114,0.08536),
		EARTH_COLOR,
		EARTH_NUMBER,
		vec(-0.0075,0.77251,0.14993),
		mat_mul(mat_scaling(0.106,0.106,0.106),
			mat_rotationAbout(vec(-0.416,0.890,-0.195),0.881)),
		0.12),
	{
		class = "ZhandulParticle",
		emitter = "linwfr_diamond",
		offset = linwfrDiamondCenter,
		rotation = vec(45,0,0),
	},
	{
		class = "ZhandulParticle",
		name = "zp2",
		emitter = "linwfr_diamond_2",
		offset = linwfrDiamondCenter,
		rotation = vec(0,30,0),
	},
}
for i=1,#linwfrKeyPositions/2 do
	local scale = linwfrKeyPositions[i*2]
	linwfrComponents[#linwfrComponents+1] = {
		class = "Model",
		name = "key"..i,
		offset = mat_transformVector(linwfrKeyPositions[i*2-1]-goddessData[EARTH_NUMBER].position,mat_rotationY(MATH_PI)),
		model = "mod_assets/linwfr_key_circle.fbx",
		material = "eye",
		emissiveColor = vec(0,1,0),
		onInit = function(self)
			self = proxyToObject(self)
			local xz = i == 3 and 0.4 or 0.6 -- HACK: j skinny
			
			-- no, really, 0.5 radians is about the right value here
			self.node:setRotation(mat_mul(mat_rotationY(0.5),mat_scaling(scale*xz,scale,scale*xz)))
		end,
	}
end
defineObject{
	name = "goddess_earth",
	components = linwfrComponents,
	placement = "floor",
}
defineObject{
	name = "goddess_air",
	components = {
		{
			class = "Model",
			model = "mod_assets/ga.fbx",
			outline = true,
			emissiveColor = AIR_COLOR,
		},
		{
			class = "Model",
			name = "h",
			model = "mod_assets/gah.fbx",
			emissiveColor = AIR_COLOR,
		},
		-- prevents going inside her arm
		{
			class = "ProjectileCollider",
			offset = vec(0.36,0.54,0.25),
			size = vec(0.2,0.2,0.2),
		},
		-- prevents going inside her hair
		{
			class = "ProjectileCollider",
			name = "c2",
			offset = vec(-0.31,0.49,-0.41),
			size = vec(0.5,0.5,0.5),
		},
		goddessBaseComponent(AIR_NUMBER),
		goddessSoundComponent(AIR_NUMBER),
		goddessBaseLabelComponent([[
AIR_NAME_LOWER
v]]),
		goddessEyeComponent(vec(0.18056,0.67483,0.18248),AIR_COLOR,0,AIR_NUMBER),
		goddessEyeComponent(vec(0.16224,0.66335,0.20138),AIR_COLOR,1,AIR_NUMBER),
		goddessSpotLightComponent(vec(0.18962,0.65591,0.2053),
			AIR_COLOR,
			AIR_NUMBER,
			vec(0.4007,0.5276,0.2547),
			mat_mul(mat_scaling(0.051,0.028,0.035),
				mat_rotationAbout(vec(-0.019,0.897,-0.443),2.04)),
			0.09)
	},
	placement = "floor",
}
defineObject{
	name = "goddess_life_death",
	components = {
		{
			class = "Model",
			model = "mod_assets/gld.fbx",
			emissiveColor = LIFE_AND_DEATH_COLOR,
			outline = true,
		},
		{
			class = "Model",
			name = "h",
			model = "mod_assets/glh.fbx",
			emissiveColor = LIFE_AND_DEATH_COLOR,
		},
		{
			class = "Model",
			name = "hd",
			model = "mod_assets/gdh.fbx",
			emissiveColor = LIFE_AND_DEATH_COLOR,
		},
		goddessBaseComponent(LIFE_NUMBER+DEATH_NUMBER),
		goddessSoundComponent(LIFE_NUMBER),
		goddessBaseLabelComponent([[

LIFE_NAME_LOWER     i
DEATH_NAME_LOWER   vi]]),
		goddessEyeComponent(vec(0.11082,0.60789,0.0753),LIFE_AND_DEATH_COLOR,0,LIFE_NUMBER+DEATH_NUMBER),
		goddessEyeComponent(vec(0.08743,0.61936,0.0895),LIFE_AND_DEATH_COLOR,1,LIFE_NUMBER+DEATH_NUMBER),
		goddessSpotLightComponent(vec(0.1085,0.647,0.115),
			LIFE_AND_DEATH_COLOR,
			LIFE_NUMBER+DEATH_NUMBER,
			vec(0.04908,0.56323,0.04338),
			mat_mul(mat_scaling(0.279,0.279,0.279),
				mat_rotationAbout(vec(-0.215,0.017,0.976),2.814)),
			0.09)
	},
	placement = "floor",
}
defineObject{
	name = "goddess_fire",
	components = {
		{
			class = "Model",
			model = "mod_assets/gf.fbx",
			outline = true,
			emissiveColor = FIRE_COLOR,
		},
		{
			class = "Model",
			name = "h",
			model = "mod_assets/gfh.fbx",
			emissiveColor = FIRE_COLOR,
		},
		goddessBaseComponent(FIRE_NUMBER),
		goddessSoundComponent(FIRE_NUMBER),
		goddessBaseLabelComponent([[
FIRE_NAME_LOWER
iii]]),
		goddessEyeComponent(vec(0.16125,0.45196,0.0027),FIRE_COLOR,0,FIRE_NUMBER),
		goddessEyeComponent(vec(0.14147,0.46514,-0.014),FIRE_COLOR,1,FIRE_NUMBER),
		goddessSpotLightComponent(vec(0.1336,0.466,0.018),
			FIRE_COLOR,
			FIRE_NUMBER,
			vec(-0.02,0.3286,0.0347),
			mat_mul(mat_scaling(0.38,0.375,0.25),
				mat_rotationAbout(vec(-0.102,-0.817,-0.567),0.661)),
			0.074)
	},
	placement = "floor",
}
defineZhandulEmitter{
	name = "goddess_final",
	emissionRate = 32768,
	sprayAngle = {0, 360},
	boxMin = {0, 0, 0.25},
	boxMax = {0, 0, 0.25},
	velocity = {0,0.5},
	lifetime = {1.0, 1.0},
	color0 = {1.0, 1.0, 1.0},
	fadeIn = 0.0,
	fadeOut = 1.0,
	gravity = {0,0,0},
	airResistance = 2.6,
	rotationSpeed = 0,
	blendMode = "Additive",
	objectSpace = true,
	shader = "particles_goddess_final",
}
defineObject{
	name = "goddess_final",
	components = {
		{
			class = "Model",
			model = "mod_assets/gfinal.fbx",
			material = "eye", -- bit of a hack but whatever
			emissiveColor = vec(-1,-1,-1),
		},
		{
			class = "Sound",
			sound = "goddess_noise_9",
		},
		{
			class = "Sound",
			name = "sound2",
			sound = "goddess_noise_9",
			pitch = MATH_PI*0.4,
		},
		{
			class = "ZhandulParticle",
			emitter = "goddess_final",
		},
		goddessSpotLightComponent(vec(0,0,0),
			FINAL_COLOR,
			0)
	},
	placement = "floor",
}

defineObject{
	name = "roomwall",
	components = {
		{
			class = "Model",
			model = "mod_assets/w.fbx",
		},
	},
	placement = "wall",
}

#define WALLSCROLL_TEXT_Z_OFFSET -0.01
#define SMALL_TEXT_SCALE 0.04
local WALL_TEXT_OFFSET = vec(0,0,WALLSCROLL_TEXT_Z_OFFSET)
local WALL_TEXT_OFFSET_SMALL = WALL_TEXT_OFFSET+vec(0.7,-0.5,0)
-- XXX: there's an obvious optimization being completely left on the table here:
-- there are plenty of situations where you can be certain a conditional text
-- model won't be visible (from which goddesses are flipped or flipping or 
-- nonexistent).
-- Profiling with all models visible vs. hidden suggests that this optimization
-- is not worth the time to write, but somebody else's computer will probably
-- prove me wrong...
defineObject{
	name = "wallscroll",
	components = {
		{
			class = "Model",
			model = "mod_assets/wallscroll.fbx",
			outline = true,
		},
		{
			class = "Model",
			name = "t000",
			offset = WALL_TEXT_OFFSET,
			text = bookTextUnlit,
			material = "conditional_text_black",
		},
		{
			class = "Model",
			name = "t100",
			emissiveColor = vec(1,0,0),
			offset = WALL_TEXT_OFFSET,
			text = bookTextRed,
			material = "conditional_text",
		},
		{
			class = "Model",
			name = "t010",
			emissiveColor = vec(0,1,0),
			offset = WALL_TEXT_OFFSET,
			text = bookTextGreen,
			material = "conditional_text",
		},
		{
			class = "Model",
			name = "t001",
			emissiveColor = vec(0,0,1),
			offset = WALL_TEXT_OFFSET,
			text = bookTextBlue,
			material = "conditional_text",
		},
		{
			class = "Model",
			name = "t110",
			emissiveColor = vec(1,1,0),
			offset = WALL_TEXT_OFFSET,
			text = bookTextYellow,
			material = "conditional_text",
		},
		{
			class = "Model",
			name = "t101",
			emissiveColor = vec(1,0,1),
			offset = WALL_TEXT_OFFSET,
			text = bookTextMagenta,
			material = "conditional_text",
		},
		{
			class = "Model",
			name = "t011",
			emissiveColor = vec(0,1,1),
			offset = WALL_TEXT_OFFSET,
			text = bookTextCyan,
			material = "conditional_text",
		},
		{
			class = "Model",
			name = "t111",
			emissiveColor = vec(1,1,1),
			offset = WALL_TEXT_OFFSET,
			text = bookTextWhite,
			material = "conditional_text",
		},
		{
			class = "Model",
			name = "t0-0",
			emissiveColor = vec(0,-1,0),
			offset = WALL_TEXT_OFFSET,
			text = bookTextNGreen,
			material = "conditional_text",
		},
		{
			class = "Model",
			name = "t00-",
			emissiveColor = vec(0,0,-1),
			offset = WALL_TEXT_OFFSET,
			text = bookTextNBlue,
			material = "conditional_text",
		},
		{
			class = "Model",
			name = "t0--",
			emissiveColor = vec(0,-1,-1),
			offset = WALL_TEXT_OFFSET,
			text = bookTextNCyan,
			material = "conditional_text",
		},
		{
			class = "Model",
			name = "t01-",
			emissiveColor = vec(0,1,-1),
			offset = WALL_TEXT_OFFSET,
			text = bookTextGreenNBlue,
			material = "conditional_text",
		},
		{
			class = "Model",
			name = "t0-1",
			emissiveColor = vec(0,-1,1),
			offset = WALL_TEXT_OFFSET,
			text = bookTextBlueNGreen,
			material = "conditional_text",
		},
		-- the next 5 are identical so i guess you could like optimize
		{
			class = "Model",
			name = "t1-0",
			emissiveColor = vec(1,-1,0),
			offset = WALL_TEXT_OFFSET,
			text = bookTextRedNGreen,
			material = "conditional_text",
		},
		{
			class = "Model",
			name = "t10-",
			emissiveColor = vec(1,0,-1),
			offset = WALL_TEXT_OFFSET,
			text = bookTextRedNBlue,
			material = "conditional_text",
		},
		{
			class = "Model",
			name = "t1--",
			emissiveColor = vec(1,-1,-1),
			offset = WALL_TEXT_OFFSET,
			text = bookTextRedNCyan,
			material = "conditional_text",
		},
		{
			class = "Model",
			name = "t11-",
			emissiveColor = vec(1,1,-1),
			offset = WALL_TEXT_OFFSET,
			text = bookTextYellowNBlue,
			material = "conditional_text",
		},
		{
			class = "Model",
			name = "t1-1",
			emissiveColor = vec(1,-1,1),
			offset = WALL_TEXT_OFFSET,
			text = bookTextMagentaNGreen,
			material = "conditional_text",
		},
		{
			class = "Model",
			name = "t---",
			emissiveColor = vec(-1,-1,-1),
			offset = WALL_TEXT_OFFSET,
			text = bookTextFinal,
			material = "conditional_text",
		},
		-- small ones
		{
			class = "Model",
			name = "t111s",
			emissiveColor = vec(1,1,1),
			offset = WALL_TEXT_OFFSET_SMALL,
			text = {bookTextWhiteSmall,scale=SMALL_TEXT_SCALE},
			material = "conditional_text_tiny",
		},
		{
			class = "Model",
			name = "t001s",
			emissiveColor = vec(0,0,1),
			offset = WALL_TEXT_OFFSET_SMALL,
			text = {bookTextBlueSmall,scale=SMALL_TEXT_SCALE},
			material = "conditional_text_tiny",
		},
		{
			class = "Model",
			name = "t010s",
			emissiveColor = vec(0,1,0),
			offset = WALL_TEXT_OFFSET_SMALL,
			text = {bookTextGreenSmall,scale=SMALL_TEXT_SCALE},
			material = "conditional_text_tiny",
		},
		{
			class = "Model",
			name = "t011s",
			emissiveColor = vec(0,1,1),
			offset = WALL_TEXT_OFFSET_SMALL,
			text = {bookTextCyanSmall,scale=SMALL_TEXT_SCALE},
			material = "conditional_text_tiny",
		},
	},
	placement = "wall",
}
defineObject{
	name = "backwalltext",
	components = {
		{
			class = "Model",
			model = "mod_assets/wallscroll.fbx",
			outline = true,
			onInit = function(self)
				proxyToObject(self).node:transform(mat_scaling(0.3,0.3,0.3))
			end
		},
		{
			class = "Model",
			name = "text",
			text = {[[
it takes more   
to make a world 
the second actor
is named irsten ]],scale=0.7},
			emissiveColor = PHYS_COLOR,
			offset = WALL_TEXT_OFFSET,
			material = "conditional_text",
		},
	},
	placement = "wall",
}

#define AZCHXA_MULT 1.1
local STAR_CHART_DATA = {
	130,  0, -- A
	228,125, -- B
	  0, 97, -- C
	190,134, -- D
	161,216, -- E
	 84,183, -- F
	 33,142, -- G
	263, 94, -- H
	109,130, -- I
	212,206, -- J
	 95, 95, -- K
	 82,145, -- L
	203, 63, -- M
	133,209, -- N
	155, 44, -- O
	 48,206, -- P
	134, 79, -- Q
	126,170, -- R
	163,150, -- S
	158,107, -- T
	206,101, -- U
	 89, 38, -- V
	 99,218, -- W
	 54,249, -- X
	 54, 76, -- Y
	213,247, -- Z
	199,171, -- *
}
local STAR_CHART_SIZE = 263
local STAR_CHART_SCALE = 8
local starChartComponents = {}
for i=1,27 do
	local char = i == 27 and "*" or string.char(i+96)
	local posScale = STAR_CHART_SCALE
	-- holy shit kill me
	-- (a last-day-development hack: moves azchxa's letters closer to the
	-- wall without distorting the pentagram, so that they're easier to pick
	-- out from the rest of the letters)
	if char == "a" or char == "c" or char == "h" or char == "x" or char == "z" then
		posScale = posScale*AZCHXA_MULT
	end
	starChartComponents[i] = {
		class = "Model",
		name = char,
		text = {char,scale=1.2},
		material = "unconditional_text",
		emissiveColor = vec(1,1,1),
		offset = vec(STAR_CHART_DATA[i*2-1]/STAR_CHART_SIZE-0.5,0.001,-STAR_CHART_DATA[i*2]/STAR_CHART_SIZE+0.5)*posScale,
		rotation = vec(270,0,0),
		enabled = false,
	}
end
defineObject{
	name = "starchart",
	components = starChartComponents,
	placement = "floor",
}

defineObject{
	name = "stars",
	components = {
		{
			class = "Model",
			enabled = false,
			onInit = function(self)
				self = proxyToObject(self)
				local vertexArray = {}
				local texcoordArray = {}
				local indices = {}
				local colorArray = {}
				
				local function addStar(x,y,z,r,g,b)
					local firstVertex = #vertexArray+1
					for vertex=1,4 do
						local vi = #vertexArray+1
						vertexArray[vi] = x
						vertexArray[vi+1] = y
						vertexArray[vi+2] = z
						local ci = #colorArray+1
						colorArray[ci] = r
						colorArray[ci+1] = g
						colorArray[ci+2] = b
						colorArray[ci+3] = 127
					end
					local ti = #texcoordArray+1
					texcoordArray[ti] = 0 -- NW
					texcoordArray[ti+1] = 1
					texcoordArray[ti+2] = 1 -- NE
					texcoordArray[ti+3] = 1
					texcoordArray[ti+4] = 0 -- SW
					texcoordArray[ti+5] = 0
					texcoordArray[ti+6] = 1 -- SE
					texcoordArray[ti+7] = 0
					
					-- Connect the vertices with two triangles
					local ii = #indices+1
					local sii = (firstVertex-1)/3 -- NW vertex
					indices[ii] = sii -- NW-NE-SW triangle
					indices[ii+1] = sii+1
					indices[ii+2] = sii+2
					indices[ii+3] = sii+3 -- NE-SW-SE triangle
					indices[ii+4] = sii+2
					indices[ii+5] = sii+1
				end
				
				local stars = 0
				local diameter = 200
				local radius = diameter/2
				local yRange = radius-3
				while stars < 32768 do
					-- pick random points in a cube, then
					-- discard those that aren't in the
					-- inscribed sphere or high enough
					local x = (math_random()-0.5)*diameter
					-- same as discarding every point where y < 3
					local y = 3+math_random()*yRange
					local z = (math_random()-0.5)*diameter
					
					if (x*x+y*y+z*z) <= radius*radius then
						addStar(x,y,z,127,127,127)
						stars = stars+1
					end
				end

				local nameStarCount = 2048
				local nameMesh = Mesh.load("mod_assets/sn2.fbx")
				local namePoints = getRandomPointsOnMesh(nameMesh, nil, nameStarCount, 8, 19)
				for i=1,nameStarCount do
					addStar(namePoints[i*3-2],namePoints[i*3-1],namePoints[i*3],64,64,127)
				end

				local mesh = Mesh.create(#vertexArray/3)
				mesh:setVertexArray(vertexArray)
				mesh:setTexcoordArray(texcoordArray)
				mesh:setIndices(indices)
				mesh:setColorArray(colorArray)
				mesh:addSegment(dungeon.materials.s.material,"triangles",0,#indices/3)
				-- we don't need normals or tangents
				local rm = RenderableMesh.create(mesh)
				local me = MeshEntity.create(rm)
				
				me:setMesh(rm)
				
				self.node.RenderEntity = me
				self.node:addTo(self.go.node)
				if not self.enabled then me:setHidden(true) end
			end,
		},
	},
	placement = "floor",
}

defineObject{
	name = "party",
	components = {
		{
			class = "Party",
			onInit = function(self)
				-- we don't use this and not having to update it saves
				-- surprisingly many microseconds
				self.go:removeComponent("gravity")
			
				local map = dungeon.maps[1]
				
				for i=1,6 do
					local pedestal = spawn(map,"pedestal",0,0,0,0,"pedestal"..i)
					pedestal:setWorldPosition(goddessPosition(2.5+i)-vec(0, PEDESTAL_HEIGHT, 0))
				end
				
				spawn(map,"floor",1,1,0,0,"floor")
				spawn(map,"ceiling",1,1,0,0,"ceiling")
				spawn(map,"roomwall",1,0,0,0,"wall_n")
				spawn(map,"roomwall",2,1,1,0,"wall_e")
				spawn(map,"roomwall",1,2,2,0,"wall_s")
				spawn(map,"roomwall",0,1,3,0,"wall_w")
				local wallscroll = spawn(map,"wallscroll",1,0,0,0,"wallscroll")
				wallscroll:setWorldPosition(wallscroll:getWorldPosition()+vec(0,1.5,-0.1))
				spawn(map,"starchart",1,1,0,0,"starchart")
				spawn(map,"backwalltext",1,1,2,0,"backwalltext")
					:setWorldPosition(vec(5.65,1.5,0.01))
				spawn(map,"stars",1,1,0,0,"stars")
				soundSystem:playStream("")
				
				if not config.editor then
					burdenOfKnowledgeLoad()
				end

				--[[
				-- debug spawn goddesses
				for i=2,7 do
					spawnGoddess(i)
				end
				-- debug unspawn ceiling
				destroyCeilingLogic()
				
				config.profiler = true
				config.developer = true
				]]
			end,
		},
	},
	editorIcon = 32,
	placement = "floor",
}

--------------------------------------------------------
-- savegame code!!!
--------------------------------------------------------
local function burdenOfKnowledgeSaveInner(data)
	local filename = config.documentsFolder.."/burden_of_knowledge.sav"
	local file = io.open(filename,"wb")
	file:write(data)
	file:close()
end
function burdenOfKnowledgeSave()
	local data = ""
	for i,invoked in pairs(invokedGoddesses) do
		if invoked then data = data..i end
	end
	if not findObject("ceiling") then
		data = data.."c"
	end
	
	local ok,err = pcall(burdenOfKnowledgeSaveInner,data)
	
	if ok then
		saveFailed = false
	else
		infoMessage("WARNING: Could not save progress.")
		saveFailed = true
		return
	end
end
function burdenOfKnowledgeLoad(dryRun)
	local filename = config.documentsFolder.."/burden_of_knowledge.sav"
	local ok,exists = pcall(FileSystem.fileExists,filename)
	if not ok or not exists then return end
	local ok,data = pcall(FileSystem.readFile,filename)
	if not ok then return end
	
	local goddesses = {}
	local killCeiling = false
	for i=1,#data do
		local c = data:sub(i,i)
		if c == "c" then
			killCeiling = true
		else
			local n = tonumber(c)
			if n and goddessData[n] then
				goddesses[n] = true
			else
				return false
			end
		end
	end
	
	if not dryRun then
		if killCeiling then destroyCeilingLogic() end
		for goddess,_ in pairs(goddesses) do
			if goddess ~= LIFE_NUMBER and goddess ~= DEATH_NUMBER then
				spawnGoddess(goddess)
			end
			invokedGoddesses[goddess] = true -- still do this for life+death!!!
		end
		local ldCount = (goddesses[LIFE_NUMBER] and 1 or 0) + (goddesses[DEATH_NUMBER] and 1 or 0)
		if ldCount == 1 then
			spawnLifeDeathHalfFX()
		elseif ldCount == 2 then
			spawnGoddess(LIFE_NUMBER)
		end
		-- if player knows how to flip goddesses, start them facing away from the scroll
		if invokedGoddesses[WATER_NUMBER] then
			for goddess,_ in pairs(goddesses) do
				if goddess ~= DEATH_NUMBER then
					doGoddessFlip(goddess, true)
				end
			end
		end
	end
	
	return true
end

hintChapter = "hint"
hintPage = 0
function setHintChapter(chapter)
	if hintChapter ~= chapter then hintPage = 0 end
	hintChapter = chapter
end
function hint()
	local hint
	if not hinted then
		hinted = true
	elseif not invokedGoddesses[PHYS_NUMBER] then
		setHintChapter(PHYS_NUMBER)
	elseif not invokedGoddesses[WATER_NUMBER] then
		setHintChapter(WATER_NUMBER)
	elseif findObject("ceiling") then
		setHintChapter("ceiling")
	elseif not invokedGoddesses[EARTH_NUMBER] then
		setHintChapter(EARTH_NUMBER)
	elseif not zoomedInOnTinyText then
		setHintChapter("zoom")
	elseif not (invokedGoddesses[FIRE_NUMBER] or invokedGoddesses[AIR_NUMBER] or hintChapter == FIRE_NUMBER or (hintChapter == "fejOrAzchxa" and hintPage == #hints[hintChapter])) then
		setHintChapter("fejOrAzchxa")
	elseif not invokedGoddesses[FIRE_NUMBER] then
		setHintChapter(FIRE_NUMBER)
	elseif not invokedGoddesses[AIR_NUMBER] then
		setHintChapter(AIR_NUMBER)
	elseif not (invokedGoddesses[LIFE_NUMBER] or invokedGoddesses[DEATH_NUMBER] or hintChapter == LIFE_NUMBER or (hintChapter == "lifeOrDeath" and hintPage == #hints[hintChapter])) then
		setHintChapter("lifeOrDeath")
	elseif not invokedGoddesses[LIFE_NUMBER] then
		setHintChapter(LIFE_NUMBER)
	elseif not invokedGoddesses[DEATH_NUMBER] then
		setHintChapter(DEATH_NUMBER)
	elseif not invokedGoddesses[FINAL_NUMBER] then
		setHintChapter(FINAL_NUMBER)
	else
		setHintChapter("final")
	end
	hintPage = math_min(hintPage+1,#hints[hintChapter])
	hint = hints[hintChapter][hintPage]
	
	if hint then infoMessage(hint) end
	return true
end
