#version 150
#include "base.h"
#include "../../mod_assets/negative_light.h"

in vec3 positionVarying;	// unused
in vec3 normalVarying;		// unused
in vec2 texcoordVarying;

out vec4 fragColor;

#define DITHER_STEPS 8
#define DITHER_TEX_STEP_SIZE 4
uniform sampler2D g_ditherTex;
//uniform sampler2D g_diffuseMap;
//uniform sampler2D g_specularMap;
//uniform sampler2D g_emissiveMap;
uniform sampler2D g_lightBuffer;
//uniform vec3 g_emissiveColor;
uniform vec2 g_invScreenSize;
uniform float g_time;
uniform float g_dissolve;

#define DITHER_THRESHOLD_FULL 1.5
#define DITHER_THRESHOLD_NONE 0.125
float ditherChannel(float intensity, ivec2 fragPixel)
{
	int step = int((intensity-DITHER_THRESHOLD_NONE)/(DITHER_THRESHOLD_FULL-DITHER_THRESHOLD_NONE)*DITHER_STEPS);
	step = clamp(step, 0, DITHER_STEPS-1);
	ivec2 coord = ivec2(step*DITHER_TEX_STEP_SIZE+fragPixel.x%4, fragPixel.y%4);
	return texelFetch(g_ditherTex, coord, 0).x;
}

vec3 ditherLight(vec3 light, ivec2 fragPixel)
{
	return vec3(ditherChannel(light.x,fragPixel),ditherChannel(light.y,fragPixel),ditherChannel(light.z,fragPixel));
}

void main()
{
	//vec4 diffuseColor = vec4(1.0,1.0,1.0,1.0); //texture(g_diffuseMap, texcoordVarying);
#ifdef ALPHA_TEST
	//if(diffuseColor.w < 0.5)
	//	discard;
#endif
	//vec3 specularColor = texture(g_specularMap, texcoordVarying).xyz;
	//vec3 emissiveColor = texture(g_emissiveMap, texcoordVarying).xyz;
	
	//vec3 normal = normalize(normalVarying);
	//vec3 incident = normalize(positionVarying.xyz);
	
	vec2 fragScreenCoord = gl_FragCoord.xy * g_invScreenSize;
	ivec2 fragPixel = ivec2(gl_FragCoord);
	vec3 light = ditherLight(resolveLight(texture(g_lightBuffer, fragScreenCoord).xyz, g_time), fragPixel);
	
	if (ditherChannel(g_dissolve, fragPixel) > 0.5) {
		discard;
	}

	fragColor = vec4(light, 1.0);
}
