#version 150
#include "base.h"

in vec4 position;
in vec3 normal;

#ifdef SKINNING
in vec4 boneIndices;
in vec4 boneWeights;
#endif

out vec3 positionVarying;
out vec3 normalVarying;

uniform mat4 g_modelView;
uniform mat4 g_proj;
uniform vec2 g_invScreenSize;

void main()
{
	float outlineWidth = 2.0;

	vec4 vposition = position;
	// mesh has flipped normals, so invert them again
	vec3 vnormal = -unpackVertexNormal(normal);

#ifdef SKINNING
	mat4 skinMat = skinningMatrix(boneIndices, boneWeights);
	vposition = vposition * skinMat;
	vnormal = vec3(vec4(vnormal, 0.0) * skinMat);
#endif

	vec4 pos = g_modelView * vposition;

	normalVarying = vec3(g_modelView * vec4(vnormal, 0.0));
	
	vec4 clipPos = g_proj * pos;
	vec3 clipNormal = (g_proj * vec4(normalVarying, 0.0)).xyz;
	
	clipPos.xy += normalize(clipNormal.xy) * g_invScreenSize * outlineWidth * clipPos.w * 2;
	
	gl_Position = clipPos;
	positionVarying = pos.xyz;
}
