#version 150
#include "base.h"
#include "../../mod_assets/shaders/gl/particles.h"

// from ParticleSystem.vsh
vec3 particlePos_airResistance(vec3 v0, vec3 g, float k, float t)
{
	return g * (t/k) + (1.0/k) * (v0 - g/k) * (1.0 - exp(-k*t));
}

void main()
{
	vec4 ncolor = color*INV_COLOR_MAX;
	vec3 nnormal = normal*(1.0/255.0);

	float cycleDuration = g_particleLifetime.z;
	float lifetime = mix(g_particleLifetime.x, g_particleLifetime.y, ncolor.y);
	float emittedTime = ncolor.z * cycleDuration;
	float endTime = emittedTime+lifetime;
	float timeWithinCycle = mod(g_particleTime, cycleDuration);
	
	float timeSinceEmission = mod(timeWithinCycle - emittedTime, cycleDuration);
	
	vec4 pos = position;
	
	colorVarying = g_particleColor;
	
	if (g_particleTime < emittedTime || // first cycle, particle hasn't reached its emission time yet
		// particle does not cross cycle boundary and is finished for this cycle
		timeWithinCycle > endTime ||
		// particle was not emitted before emitter was stopped
		timeSinceEmission < g_timeSinceStop ||
		// all other inactivity cases
		(timeWithinCycle < emittedTime && (endTime < cycleDuration || timeWithinCycle > mod(endTime, cycleDuration)))) {
		
		// Inactive particle, put it somewhere it'll definitely get culled.
		gl_Position = vec4(-10000, -10000, 0, 1);
	} else {
		// get initial velocity
		// if you're reading this comment it means i didn't come back and replace
		// the easy-to-understand spherical coordinate test code with something that actually
		// performs well.
		float azimuthal = nnormal.x * M_PI * 2.0;
		vec2 sprayAngle = g_particleMovementParams.zw;
		float polar = mix(sprayAngle.x, sprayAngle.y, nnormal.y);
		vec2 velocityParams = g_particleMovementParams.xy;
		vec3 dir = vec3(sin(polar)*cos(azimuthal), sin(polar)*sin(azimuthal), cos(polar));
		vec3 velocity = dir *
			mix(velocityParams.x, velocityParams.y, nnormal.z);
			
		pos.xyz = dir*g_particleBoxMax.z;
		
		// now we can stop worrying about cycle crap and look more like a respectable particle shader.
		// Well, maybe not respectable, more like "loosely tolerable considering it's a game mod with an
		// experimental aesthetic where efficiency isn't a big priority and it probably took less than a
		// month to make."
		pos.xyz += particlePos_airResistance(velocity, g_particleGravity, g_particleAirResistance, timeSinceEmission);
		
		// handle fade, same as vanilla particle fade except multiplies color instead of alpha
		float invFadeIn = g_particleFadeParams.x;
		float invFadeOut = g_particleFadeParams.y;
		float fadeOut = g_particleFadeParams.z;
		float startFadeOut = lifetime - fadeOut;
		float fadeMult = saturate(timeSinceEmission * invFadeIn) * saturate(1.0 - (timeSinceEmission - startFadeOut) * invFadeOut);

		pos = g_modelView * pos;
		
		pos.xy = pos.xy + particleVertexExpansionDirection()*0.004*fadeMult;
		
		pos = g_proj * pos;
		
		gl_Position = pos;
	}
	positionVarying = pos.xyz;
}
