#version 150
#include "base.h"
#include "../../mod_assets/shaders/gl/particles.h"

// from ParticleSystem.vsh
vec3 particlePos_airResistance(vec3 v0, vec3 g, float k, float t)
{
	return g * (t/k) + (1.0/k) * (v0 - g/k) * (1.0 - exp(-k*t));
}

void main()
{
	vec4 ncolor = color*INV_COLOR_MAX;
	vec3 nnormal = normal*(1.0/255.0);

	float lifetime = mix(g_particleLifetime.x, g_particleLifetime.y, ncolor.x);
	
	vec4 pos = position;
	
	colorVarying = g_particleColor;
	
	if (g_particleTime > lifetime) {
		// Inactive particle, put it somewhere it'll definitely get culled.
		gl_Position = vec4(-10000, -10000, 0, 1);
	} else {
		// map position to box; this gives us the particle's initial position
		pos.xyz = pos.xyz * g_particleBoxMin + (vec3(1.0,1.0,1.0) - pos.xyz) * g_particleBoxMax;
		
		// lazy as hell
		vec2 velocityParams = g_particleMovementParams.xy;
		vec3 velocity = vec3(sin(ncolor.y*(M_PI*2.0)), cos(ncolor.y*(M_PI*2.0)), 0) *
			mix(velocityParams.x, velocityParams.y, ncolor.z);
		
		pos.xyz += particlePos_airResistance(velocity, g_particleGravity, g_particleAirResistance, g_particleTime);
		
		// This is where you'd normally rotate and resize the particle, but we just want a pixel, so we're
		// not doing that.
		
		// transform to clip space
		pos = g_modelView * pos;
		pos = g_proj * pos;
		
		// expand to pixel-size regardless of distance
		pos.xy = pos.xy + particleVertexExpansionDirection()*g_invScreenSize*pos.w*2.0;
		gl_Position = pos;
		
		// handle fade, same as vanilla particle fade except multiplies color instead of alpha
		float invFadeIn = g_particleFadeParams.x;
		float invFadeOut = g_particleFadeParams.y;
		float fadeOut = g_particleFadeParams.z;
		float startFadeOut = lifetime - fadeOut;
		colorVarying.xyz *= saturate(g_particleTime * invFadeIn) * saturate(1.0 - (g_particleTime - startFadeOut) * invFadeOut);
	}
	positionVarying = pos.xyz;
}
