#version 150
#include "base.h"
#include "../../mod_assets/shaders/gl/particles.h"

// from ParticleSystem.vsh
vec3 particlePos_airResistance(vec3 v0, vec3 g, float k, float t)
{
	return g * (t/k) + (1.0/k) * (v0 - g/k) * (1.0 - exp(-k*t));
}

void main()
{
	vec4 ncolor = color*INV_COLOR_MAX;
	vec3 nnormal = normal*(1.0/255.0);

	float cycleDuration = g_particleLifetime.z;
	float lifetime = mix(g_particleLifetime.x, g_particleLifetime.y, ncolor.x);
	float emittedTime = ncolor.y * cycleDuration;
	float endTime = emittedTime+lifetime;
	float timeWithinCycle = mod(g_particleTime, cycleDuration);
	
	vec4 pos = position;
	
	if (g_particleTime < emittedTime || // first cycle, particle hasn't reached its emission time yet
		// particle does not cross cycle boundary and is finished for this cycle
		timeWithinCycle > endTime ||
		// all other inactivity cases
		(timeWithinCycle < emittedTime && (endTime < cycleDuration || timeWithinCycle > mod(endTime, cycleDuration)))) {
		
		// Inactive particle, put it somewhere it'll definitely get culled.
		gl_Position = vec4(-10000, -10000, 0, 1);
	} else {
		// map position to box; this gives us the particle's initial position
		pos.xyz = pos.xyz * g_particleBoxMin + (vec3(1.0,1.0,1.0) - pos.xyz) * g_particleBoxMax;
		
		// get initial velocity
		// if you're reading this comment it means i didn't come back and replace
		// the easy-to-understand spherical coordinate test code with something that actually
		// performs well.
		float azimuthal = nnormal.x * M_PI * 2.0;
		vec2 sprayAngle = g_particleMovementParams.zw;
		float polar = mix(sprayAngle.x, sprayAngle.y, nnormal.y);
		vec2 velocityParams = g_particleMovementParams.xy;
		vec3 velocity = vec3(sin(polar)*cos(azimuthal), sin(polar)*sin(azimuthal), cos(polar)) *
			mix(velocityParams.x, velocityParams.y, nnormal.z);
			
		float timeSinceEmission = mod(timeWithinCycle - emittedTime, cycleDuration);
		
		// now we can stop worrying about cycle crap and look more like a respectable particle shader.
		// Well, maybe not respectable, more like "loosely tolerable considering it's a game mod with an
		// experimental aesthetic where efficiency isn't a big priority and it probably took less than a
		// month to make."
		pos.xyz += particlePos_airResistance(velocity, g_particleGravity, g_particleAirResistance, timeSinceEmission);
	
		// transform to clip space
		pos = g_modelView * pos;
		pos = g_proj * pos;
		
		// expand to pixel-size regardless of distance
		pos.xy = pos.xy + particleVertexExpansionDirection()*g_invScreenSize*pos.w*2.0;
		gl_Position = pos;
		// TODO: depth bias
	}
	// TODO: fade
	float timeSinceEmission = mod(timeWithinCycle - emittedTime, cycleDuration);
	positionVarying = pos.xyz;
	float remainingTimeNormalized = (lifetime-timeSinceEmission)/lifetime;
	colorVarying = vec4(nnormal.xyz,1.0);
}
