-- Makes the mat into the identity matrix
function mat.reset(mat)
	local row = mat.x
	row[1] = 1
	row[2] = 0
	row[3] = 0
	row[4] = 0
	row = mat.y
	row[1] = 0
	row[2] = 1
	row[3] = 0
	row[4] = 0
	row = mat.z
	row[1] = 0
	row[2] = 0
	row[3] = 1
	row[4] = 0
	row = mat.w
	row[1] = 0
	row[2] = 0
	row[3] = 0
	row[4] = 1
end

-- for when working with actual vec()s would churn too much memory
function mat.transformVectorNum(v1,v2,v3,matrix)
	local mx = matrix.x
	local my = matrix.y
	local mz = matrix.z
	return v1*mx[1]+v2*my[1]+v3*mz[1],
		v1*mx[2]+v2*my[2]+v3*mz[2],
		v1*mx[3]+v2*my[3]+v3*mz[3]
end

function mat.scale(mat, scaleX, scaleY, scaleZ)
	local rval = mat.scaling(scaleX, scaleY, scaleZ)
	return mat*rval
end

function mat.scaleInPlace(mat, scaleX, scaleY, scaleZ)
	local row = mat.x
	row[1] = row[1]*scaleX
	row = mat.y
	row[2] = row[2]*scaleY
	row = mat.z
	row[3] = row[3]*scaleZ
end

function mat.mulInPlace(mat1,mat2)
	local mat1x = mat1.x
	local mat1y = mat1.y
	local mat1z = mat1.z
	local mat1w = mat1.w
	
	local mat1x1 = mat1x[1]
	local mat1x2 = mat1x[2]
	local mat1x3 = mat1x[3]
	local mat1x4 = mat1x[4]

	local mat1y1 = mat1y[1]
	local mat1y2 = mat1y[2]
	local mat1y3 = mat1y[3]
	local mat1y4 = mat1y[4]
	
	local mat1z1 = mat1z[1]
	local mat1z2 = mat1z[2]
	local mat1z3 = mat1z[3]
	local mat1z4 = mat1z[4]
	
	local mat1w1 = mat1w[1]
	local mat1w2 = mat1w[2]
	local mat1w3 = mat1w[3]
	local mat1w4 = mat1w[4]
	
	local mat2x = mat2.x
	local mat2y = mat2.y
	local mat2z = mat2.z
	local mat2w = mat2.w

	for i = 1,4 do
		local mat2xi = mat2x[i]
		local mat2yi = mat2y[i]
		local mat2zi = mat2z[i]
		local mat2wi = mat2w[i]
		mat1x[i] = mat1x1 * mat2xi
			+ mat1x2 * mat2yi
			+ mat1x3 * mat2zi
			+ mat1x4 * mat2wi
		mat1y[i] = mat1y1 * mat2xi
			+ mat1y2 * mat2yi
			+ mat1y3 * mat2zi
			+ mat1y4 * mat2wi
		mat1z[i] = mat1z1 * mat2xi
			+ mat1z2 * mat2yi
			+ mat1z3 * mat2zi
			+ mat1z4 * mat2wi
		mat1w[i] = mat1w1 * mat2xi
			+ mat1w2 * mat2yi
			+ mat1w3 * mat2zi
			+ mat1w4 * mat2wi
	end
end

-- Static!
function mat.getLookAtMatrix(origin, target, up)
	up = up or {0,1,0}
	
	-- z = normalize(target-origin)
	local offsetX = target[1]-origin[1]
	local offsetY = target[2]-origin[2]
	local offsetZ = target[3]-origin[3]
	local denom = (offsetX*offsetX+offsetY*offsetY+offsetZ*offsetZ)^0.5
	local rval = mat.create()
	local rowZ = rval.z
	rowZ[1] = offsetX/denom
	rowZ[2] = offsetY/denom
	rowZ[3] = offsetZ/denom
	rowZ[4] = 0
	
	-- x = normalize(cross(up, z))
	local vx = up[2]*target[3]-up[3]*target[2]
	local vy = up[3]*target[1]-up[1]*target[3]
	local vz = up[1]*target[2]-up[2]*target[1]
	local denom = (vx*vx+vy*vy+vz*vz)^0.5
	local rowX = rval.x
	rowX[1] = vx/denom
	rowX[2] = vy/denom
	rowX[3] = vz/denom
	rowX[4] = 0
	
	-- y = cross(z,x)
	local rowY = rval.y
	rowY[1] = rowZ[2]*rowX[3]-rowZ[3]*rowX[2]
	rowY[2] = rowZ[3]*rowX[1]-rowZ[1]*rowX[3]
	rowY[3] = rowZ[1]*rowX[2]-rowZ[2]*rowX[1]
	rowY[4] = 0
	
	local rowW = rval.w
	rowW[1] = 0
	rowW[2] = 0
	rowW[3] = 0
	rowW[4] = 1
	
	return rval
end

-- Static! 
function mat.getLookAtAndScaleMatrix(origin, target, uniformScale)
	local dx = target[1]-origin[1]
	local dy = target[2]-origin[2]
	local dz = target[3]-origin[3]

	local mat = mat.identity()
	local matX = mat.x
	local matY = mat.y
	local matZ = mat.z
	if dx < 0.0001 and dx > -0.0001 and dz < 0.0001 and dz > -0.0001 then
		matY[1] = 0
		matX[2] = 0
		matX[3] = 0
		if dy < 0.0001 and dy > 0.0001 then
			-- origin and target are at nearly the same location
			-- scale to nearly 0 and don't care about rotation
			matX[1] = 0.0001
			matY[2] = 0.0001
			matY[3] = 0
			matZ[2] = 0
			matZ[3] = 0.0001
		else
			-- object is directly above or below the target, can't use usual formula
			matX[1] = 1
			matY[2] = 0
			matY[3] = dy > 0 and -1 or 1

			matZ[2] = dy
			matZ[3] = 0
		end
		matZ[1] = 0
	else
		local length =(dx^2+dy^2+dz^2)^0.5
		local invDist = 1/length
		local zaxis1 = dx*invDist
		local zaxis2 = dy*invDist
		local zaxis3 = dz*invDist
		
		local xaxis1 = -zaxis3
		local xaxis3 = zaxis1
		local xaxislen = (xaxis1^2+xaxis3^2)^0.5
		xaxis1 = xaxis1/xaxislen
		xaxis3 = xaxis3/xaxislen
		
		local xyscale = uniformScale and length or 1
		matX[1] = xaxis1*xyscale
		matX[2] = 0
		matX[3] = xaxis3*xyscale
		matY[1] = -xaxis3*zaxis2*xyscale
		matY[2] = (xaxis3*zaxis1-xaxis1*zaxis3)*xyscale
		matY[3] = xaxis1*zaxis2*xyscale
		-- scale the z by the length
		matZ[1] = zaxis1*length
		matZ[2] = zaxis2*length
		matZ[3] = zaxis3*length
	end
	return mat
end

-- mostly naive translation from mesa C code
-- could make a version of this like mat.invertFast that assumes w is (0,0,0,1)
-- and x[4],y[4],z[4] are 0
function mat.invertInPlace(mat)
	local x = mat.x
	local y = mat.y
	local z = mat.z
	local w = mat.w
	
	local i0 = x[1]
	local i1 = x[2]
	local i2 = x[3]
	local i3 = x[4]
	
	local i4 = y[1]
	local i5 = y[2]
	local i6 = y[3]
	local i7 = y[4]
	
	local i8 = z[1]
	local i9 = z[2]
	local i10 = z[3]
	local i11 = z[4]
	
	local i12 = w[1]
	local i13 = w[2]
	local i14 = w[3]
	local i15 = w[4]

	local outx1 = i5*i10*i15 - i5*i11*i14 - i9*i6*i15 + i9*i7*i14 + i13*i6*i11 - i13*i7*i10
	local outy1 = i4*i11*i14 - i4*i10*i15 + i8*i6*i15 - i8*i7*i14 - i12*i6*i11 + i12*i7*i10
	local outz1 = i4*i9*i15  - i4*i11*i13 - i8*i5*i15 + i8*i7*i13 + i12*i5*i11 - i12*i7*i9
	local outw1 = i4*i10*i13 - i4*i9*i14  + i8*i5*i14 - i8*i6*i13 - i12*i5*i10 + i12*i6*i9

	local determinant = i0*outx1 + i1*outy1 + i2*outz1 + i3 * outw1
	if determinant*determinant <= 0.000000000001 then
		console:warn("tried to invert a singular matrix, wrong transforms ahoy!")
		determinant = 1
	end
	determinant = 1/determinant

	x[1] = outx1*determinant
	x[2] = (i1*i11*i14 - i1*i10*i15 + i9*i2*i15 - i9*i3*i14 - i13*i2*i11 + i13*i3*i10)*determinant
	x[3] = (i1*i6*i15  - i1*i7*i14  - i5*i2*i15 + i5*i3*i14 + i13*i2*i7  - i13*i3*i6)*determinant
	x[4] = (i1*i7*i10  - i1*i6*i11  + i5*i2*i11 - i5*i3*i10 - i9*i2*i7   + i9*i3*i6)*determinant

	y[1] = outy1*determinant
	y[2] = (i0*i10*i15 - i0*i11*i14 - i8*i2*i15 + i8*i3*i14 + i12*i2*i11 - i12*i3*i10)*determinant
	y[3] = (i0*i7*i14  - i0*i6*i15  + i4*i2*i15 - i4*i3*i14 - i12*i2*i7  + i12*i3*i6)*determinant
	y[4] = (i0*i6*i11  - i0*i7*i10  - i4*i2*i11 + i4*i3*i10 + i8*i2*i7   - i8*i3*i6)*determinant

	z[1] = outz1*determinant
	z[2] = (i0*i11*i13 - i0*i9*i15  + i8*i1*i15 - i8*i3*i13 - i12*i1*i11 + i12*i3*i9)*determinant
	z[3] = (i0*i5*i15  - i0*i7*i13  - i4*i1*i15 + i4*i3*i13 + i12*i1*i7  - i12*i3*i5)*determinant
	z[4] = (i0*i7*i9   - i0*i5*i11  + i4*i1*i11 - i4*i3*i9  - i8*i1*i7   + i8*i3*i5)*determinant

	w[1] = outw1*determinant
	w[2] = (i0*i9*i14  - i0*i10*i13 - i8*i1*i14 + i8*i2*i13 + i12*i1*i10 - i12*i2*i9)*determinant
	w[3] = (i0*i6*i13  - i0*i5*i14  + i4*i1*i14 - i4*i2*i13 - i12*i1*i6  + i12*i2*i5)*determinant
	w[4] = (i0*i5*i10  - i0*i6*i9   - i4*i1*i10 + i4*i2*i9  + i8*i1*i6   - i8*i2*i5)*determinant
end

function mat.toDirectionVector(matrix)
	return vec(mat.transformVectorNum(0,0,1,matrix))
end
-- Static!
function mat.fromDirectionVector(vector)
	return mat.getLookAtMatrix(vec(), vector, nil)
end

